import asyncio
from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd
from bidict import bidict

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.derivative.backpack_perpetual import (
    backpack_perpetual_constants as CONSTANTS,
    backpack_perpetual_utils as utils,
    backpack_perpetual_web_utils as web_utils,
)
from hummingbot.connector.derivative.backpack_perpetual.backpack_perpetual_api_order_book_data_source import (
    BackpackPerpetualAPIOrderBookDataSource,
)
from hummingbot.connector.derivative.backpack_perpetual.backpack_perpetual_api_user_stream_data_source import (
    BackpackPerpetualAPIUserStreamDataSource,
)
from hummingbot.connector.derivative.backpack_perpetual.backpack_perpetual_auth import BackpackPerpetualAuth
from hummingbot.connector.derivative.position import Position
from hummingbot.connector.perpetual_derivative_py_base import PerpetualDerivativePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair, get_new_numeric_client_order_id
from hummingbot.core.data_type.common import OrderType, PositionAction, PositionMode, PositionSide, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import AddedToCostTradeFee, TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.event.events import AccountEvent, PositionModeChangeEvent
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.utils.tracking_nonce import NonceCreator
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class BackpackPerpetualDerivative(PerpetualDerivativePyBase):
    UPDATE_ORDER_STATUS_MIN_INTERVAL = 10.0
    web_utils = web_utils
    _VALID_EVENT_TYPES = {
        "orderAccepted",
        "orderCancelled",
        "orderExpired",
        "orderFill",
        "orderModified",
        "triggerPlaced",
        "triggerFailed",
        "positionOpened",
        "positionClosed",
        "positionAdjusted",
    }

    def __init__(self,
                 backpack_api_key: str,
                 backpack_api_secret: str,
                 balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
                 rate_limits_share_pct: Decimal = Decimal("100"),
                 trading_pairs: Optional[List[str]] = None,
                 trading_required: bool = True,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN,
                 ):
        self.api_key = backpack_api_key
        self.secret_key = backpack_api_secret
        self._domain = domain
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._last_trades_poll_backpack_timestamp = 1.0
        self._nonce_creator = NonceCreator.for_milliseconds()
        self._leverage = None  # Will be fetched on first use
        self._leverage_initialized = False
        self._position_mode = None
        super().__init__(balance_asset_limit, rate_limits_share_pct)
        # Backpack does not provide balance updates through websocket, use REST polling instead
        self.real_time_balance_update = False

    @staticmethod
    def backpack_order_type(order_type: OrderType) -> str:
        return "Limit" if order_type in [OrderType.LIMIT, OrderType.LIMIT_MAKER] else "Market"

    @staticmethod
    def to_hb_order_type(backpack_type: str) -> OrderType:
        return OrderType[backpack_type]

    @property
    def authenticator(self):
        return BackpackPerpetualAuth(
            api_key=self.api_key,
            secret_key=self.secret_key,
            time_provider=self._time_synchronizer)

    @property
    def name(self) -> str:
        if self._domain == "exchange":
            return "backpack_perpetual"
        else:
            return f"backpack_perpetual_{self._domain}"

    @property
    def rate_limits_rules(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self):
        return self._domain

    @property
    def client_order_id_max_length(self):
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self):
        return CONSTANTS.HBOT_ORDER_ID_PREFIX

    @property
    def trading_rules_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_PATH_URL

    @property
    def trading_pairs_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_PATH_URL

    @property
    def check_network_request_path(self):
        return CONSTANTS.PING_PATH_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    def supported_order_types(self):
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.MARKET]

    def buy(self, trading_pair: str, amount: Decimal, order_type=OrderType.LIMIT, price: Decimal = s_decimal_NaN, **kwargs) -> str:
        """
        Override to use simple uint32 order IDs for Backpack
        """
        new_order_id = get_new_numeric_client_order_id(nonce_creator=self._nonce_creator,
                                                       max_id_bit_count=CONSTANTS.MAX_ORDER_ID_LEN)
        numeric_order_id = str(new_order_id)

        safe_ensure_future(
            self._create_order(
                trade_type=TradeType.BUY,
                order_id=numeric_order_id,
                trading_pair=trading_pair,
                amount=amount,
                order_type=order_type,
                price=price,
                **kwargs,
            )
        )
        return numeric_order_id

    def sell(self, trading_pair: str, amount: Decimal, order_type: OrderType = OrderType.LIMIT, price: Decimal = s_decimal_NaN, **kwargs) -> str:
        """
        Override to use simple uint32 order IDs for Backpack
        """
        new_order_id = get_new_numeric_client_order_id(nonce_creator=self._nonce_creator,
                                                       max_id_bit_count=CONSTANTS.MAX_ORDER_ID_LEN)
        numeric_order_id = str(new_order_id)
        safe_ensure_future(
            self._create_order(
                trade_type=TradeType.SELL,
                order_id=numeric_order_id,
                trading_pair=trading_pair,
                amount=amount,
                order_type=order_type,
                price=price,
                **kwargs,
            )
        )
        return numeric_order_id

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        request_description = str(request_exception)

        is_time_synchronizer_related = (
            "INVALID_CLIENT_REQUEST" in request_description
            and (
                "timestamp" in request_description.lower()
                or "Invalid timestamp" in request_description
                or "Request has expired" in request_description
            )
        )
        return is_time_synchronizer_related

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        return str(CONSTANTS.ORDER_NOT_EXIST_ERROR_CODE) in str(
            status_update_exception
        ) and CONSTANTS.ORDER_NOT_EXIST_MESSAGE in str(status_update_exception)

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        return str(CONSTANTS.UNKNOWN_ORDER_ERROR_CODE) in str(
            cancelation_exception
        ) and CONSTANTS.UNKNOWN_ORDER_MESSAGE in str(cancelation_exception)

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            domain=self._domain,
            auth=self._auth)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return BackpackPerpetualAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            domain=self.domain,
            api_factory=self._web_assistants_factory)

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return BackpackPerpetualAPIUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 position_action: PositionAction = PositionAction.NIL,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:
        is_maker = order_type in [OrderType.LIMIT, OrderType.LIMIT_MAKER]
        return AddedToCostTradeFee(percent=self.estimate_fee_pct(is_maker))

    def exchange_symbol_associated_to_pair(self, trading_pair: str) -> str:
        return trading_pair.replace("-", "_") + "_PERP"

    def trading_pair_associated_to_exchange_symbol(self, symbol: str) -> str:
        return symbol.replace("_", "-").replace("-PERP", "")

    async def _place_order(self,
                           order_id: str,
                           trading_pair: str,
                           amount: Decimal,
                           trade_type: TradeType,
                           order_type: OrderType,
                           price: Decimal,
                           position_action: PositionAction = PositionAction.NIL,
                           **kwargs) -> Tuple[str, float]:
        amount_str = f"{amount:f}"
        order_type_enum = self.backpack_order_type(order_type)
        side_str = CONSTANTS.SIDE_BUY if trade_type is TradeType.BUY else CONSTANTS.SIDE_SELL
        symbol = self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        data = {
            "instruction": "orderExecute",
            "symbol": symbol,
            "side": side_str,
            "quantity": amount_str,
            "clientId": int(order_id),
            "orderType": order_type_enum,
        }
        if order_type_enum == "Limit":
            price_str = f"{price:f}"
            data["price"] = price_str
            data["postOnly"] = order_type == OrderType.LIMIT_MAKER
            data["timeInForce"] = CONSTANTS.TIME_IN_FORCE_GTC
        try:
            order_result = await self._api_post(
                path_url=CONSTANTS.ORDER_PATH_URL,
                data=data,
                is_auth_required=True)
            o_id = str(order_result["id"])
            transact_time = order_result["createdAt"] * 1e-3
        except IOError as e:
            error_description = str(e)

            is_post_only_rejection = (
                order_type == OrderType.LIMIT_MAKER
                and "INVALID_ORDER" in error_description
                and "Order would immediately match and take" in error_description
            )

            if is_post_only_rejection:
                side = "BUY" if trade_type is TradeType.BUY else "SELL"
                self.logger().warning(
                    f"LIMIT_MAKER {side} order for {trading_pair} rejected: "
                    f"Order price {price} would immediately match and take liquidity. "
                    f"LIMIT_MAKER orders can only be placed as maker orders (post-only). "
                    f"Try adjusting your price to ensure the order is not immediately executable."
                )
                raise ValueError(
                    f"LIMIT_MAKER order would immediately match and take liquidity. "
                    f"Price {price} crosses the spread for {side} order on {trading_pair}."
                ) from e

            # Check for server overload
            is_server_overloaded = (
                "503" in error_description
                and "Unknown error, please check your request or try again later." in error_description
            )
            if is_server_overloaded:
                o_id = "UNKNOWN"
                transact_time = self._time_synchronizer.time()
            else:
                raise
        return o_id, transact_time

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        symbol = self.exchange_symbol_associated_to_pair(trading_pair=tracked_order.trading_pair)
        api_params = {
            "instruction": "orderCancel",
            "symbol": symbol,
            "clientId": int(order_id),
        }
        cancel_result = await self._api_delete(
            path_url=CONSTANTS.ORDER_PATH_URL,
            data=api_params,
            is_auth_required=True)
        if cancel_result.get("status") == "Cancelled":
            return True
        return False

    async def _format_trading_rules(self, exchange_info_dict: List[Dict[str, Any]]) -> List[TradingRule]:
        """
        Signature type modified from dict to list due to the new exchange info format.
        """
        trading_pair_rules = exchange_info_dict.copy()
        retval = []
        for rule in trading_pair_rules:
            if not utils.is_exchange_information_valid(rule):
                continue
            try:
                trading_pair = self.trading_pair_associated_to_exchange_symbol(symbol=rule.get("symbol"))
                filters = rule.get("filters")

                min_order_size = Decimal(filters["quantity"]["minQuantity"])
                tick_size = Decimal(filters["price"]["tickSize"])
                step_size = Decimal(filters["quantity"]["stepSize"])
                min_notional = Decimal("0")  # same as Bybit inverse, disables notional validation
                retval.append(
                    TradingRule(trading_pair,
                                min_order_size=min_order_size,
                                min_price_increment=Decimal(tick_size),
                                min_base_amount_increment=Decimal(step_size),
                                min_notional_size=Decimal(min_notional)))
            except Exception:
                self.logger().exception(f"Error parsing the trading pair rule {rule}. Skipping.")
        return retval

    async def _update_trading_fees(self):
        pass

    async def _user_stream_event_listener(self):
        async for event_message in self._iter_user_event_queue():
            try:
                if not self._validate_event_message(event_message):
                    continue
                stream = event_message["stream"]
                if "positionUpdate" in stream:
                    await self._parse_and_process_position_message(event_message)
                elif "orderUpdate" in stream:
                    self._parse_and_process_order_message(event_message)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error("Unexpected error in user stream listener loop.", exc_info=True)
                await self._sleep(5.0)

    def _validate_event_message(self, event_message) -> bool:
        stream = event_message.get("stream")
        data = event_message.get("data")
        return bool(stream and data)

    async def _parse_and_process_position_message(self, event_message: Dict[str, Any]):
        data = event_message.get("data")
        hb_trading_pair = self.trading_pair_associated_to_exchange_symbol(data.get("s"))
        quantity = Decimal(data.get("q", "0"))
        side = PositionSide.LONG if quantity > 0 else PositionSide.SHORT
        position = self._perpetual_trading.get_position(hb_trading_pair, side)
        if position is not None:
            amount = abs(quantity)
            if amount == Decimal("0"):
                pos_key = self._perpetual_trading.position_key(hb_trading_pair, side)
                self._perpetual_trading.remove_position(pos_key)
            else:
                position.update_position(position_side=side,
                                         unrealized_pnl=Decimal(data["P"]),
                                         entry_price=Decimal(data["B"]),
                                         amount=amount)
        else:
            await self._update_positions()

    def _parse_and_process_order_message(self, event_message: Dict[str, Any]):
        data = event_message.get("data")
        event_type = data.get("e")
        exchange_order_id = str(data.get("i"))
        client_order_id = str(data.get("c"))

        if event_type not in self._VALID_EVENT_TYPES:
            return

        # 1) Resolve tracked order
        tracked_order = None

        if client_order_id is not None:
            tracked_order = self._order_tracker.all_updatable_orders.get(client_order_id)

        # Fallback: sometimes 'c' is absent; match by exchange_order_id
        if tracked_order is None and exchange_order_id is not None:
            for o in self._order_tracker.all_updatable_orders.values():
                if str(o.exchange_order_id) == exchange_order_id:
                    tracked_order = o
                    client_order_id = o.client_order_id  # recover internal id
                    break

        # If still not found, nothing to update
        if tracked_order is None or client_order_id is None:
            return

        # 2) Trade fill event
        if event_type == "orderFill":
            # Trade fields are only present on orderFill events
            fee_token = data.get("N")
            fee_amount = data.get("n")

            fee = TradeFeeBase.new_perpetual_fee(
                fee_schema=self.trade_fee_schema(),
                position_action=PositionAction.NIL,
                percent_token=fee_token,
                flat_fees=(
                    [TokenAmount(amount=Decimal(str(fee_amount)), token=str(fee_token))]
                    if fee_token is not None and fee_amount is not None
                    else []
                ),
            )

            fill_qty = Decimal(str(data["l"]))
            fill_price = Decimal(str(data["L"]))

            trade_update = TradeUpdate(
                trade_id=str(data["t"]),
                client_order_id=client_order_id,
                exchange_order_id=exchange_order_id,
                trading_pair=tracked_order.trading_pair,
                fee=fee,
                fill_base_amount=fill_qty,
                fill_quote_amount=fill_qty * fill_price,
                fill_price=fill_price,
                # Backpack timestamps are microseconds
                fill_timestamp=data["T"] * 1e-6,
            )
            self._order_tracker.process_trade_update(trade_update)

        # 3) Order state update
        raw_state = data.get("X")
        new_state = CONSTANTS.ORDER_STATE.get(raw_state, OrderState.FAILED)

        order_update = OrderUpdate(
            trading_pair=tracked_order.trading_pair,
            # Backpack event time is microseconds
            update_timestamp=data["E"] * 1e-6,
            new_state=new_state,
            client_order_id=client_order_id,
            exchange_order_id=exchange_order_id,
        )
        self._order_tracker.process_order_update(order_update=order_update)

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        if order.exchange_order_id is not None:
            exchange_order_id = order.exchange_order_id
            trading_pair = self.exchange_symbol_associated_to_pair(trading_pair=order.trading_pair)
            try:
                params = {
                    "instruction": "fillHistoryQueryAll",
                    "symbol": trading_pair,
                    "orderId": exchange_order_id
                }
                all_fills_response = await self._api_get(
                    path_url=CONSTANTS.MY_TRADES_PATH_URL,
                    params=params,
                    is_auth_required=True)

                # Check for error responses from the exchange
                if isinstance(all_fills_response, dict) and "code" in all_fills_response:
                    code = all_fills_response["code"]
                    if code == "INVALID_ORDER":
                        # Order doesn't exist on exchange, mark as failed
                        order_update = OrderUpdate(
                            trading_pair=order.trading_pair,
                            new_state=OrderState.FAILED,
                            client_order_id=order.client_order_id,
                            exchange_order_id=order.exchange_order_id,
                            update_timestamp=self._time_synchronizer.time(),
                            misc_updates={
                                "error_type": "INVALID_ORDER",
                                "error_message": all_fills_response.get("msg", "Order does not exist on exchange")
                            }
                        )
                        self._order_tracker.process_order_update(order_update=order_update)
                        return trade_updates

                # Process trade fills
                for trade in all_fills_response:
                    exchange_order_id = str(trade["orderId"])
                    fee = TradeFeeBase.new_perpetual_fee(
                        fee_schema=self.trade_fee_schema(),
                        position_action=PositionAction.NIL,
                        percent_token=trade["feeSymbol"],
                        flat_fees=[TokenAmount(amount=Decimal(trade["fee"]), token=trade["feeSymbol"])]
                    )
                    trade_update = TradeUpdate(
                        trade_id=str(trade["tradeId"]),
                        client_order_id=order.client_order_id,
                        exchange_order_id=exchange_order_id,
                        trading_pair=trading_pair,
                        fee=fee,
                        fill_base_amount=Decimal(trade["quantity"]),
                        fill_quote_amount=Decimal(trade["quantity"]) * Decimal(trade["price"]),
                        fill_price=Decimal(trade["price"]),
                        fill_timestamp=pd.Timestamp(trade["timestamp"]).timestamp(),
                    )
                    trade_updates.append(trade_update)
            except IOError as ex:
                if not self._is_request_exception_related_to_time_synchronizer(request_exception=ex):
                    raise
        return trade_updates

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        trading_pair = self.exchange_symbol_associated_to_pair(trading_pair=tracked_order.trading_pair)
        updated_order_data = await self._api_get(
            path_url=CONSTANTS.ORDER_PATH_URL,
            params={
                "instruction": "orderQuery",
                "symbol": trading_pair,
                "clientId": tracked_order.client_order_id},
            is_auth_required=True)

        new_state = CONSTANTS.ORDER_STATE[updated_order_data["status"]]

        order_update = OrderUpdate(
            client_order_id=tracked_order.client_order_id,
            exchange_order_id=str(updated_order_data["id"]),
            trading_pair=tracked_order.trading_pair,
            update_timestamp=updated_order_data["createdAt"] * 1e-3,
            new_state=new_state,
        )

        return order_update

    async def _update_balances(self):
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        account_info = await self._api_get(
            path_url=CONSTANTS.BALANCE_PATH_URL,
            params={"instruction": "balanceQuery"},
            is_auth_required=True)

        if account_info:
            for asset_name, balance_entry in account_info.items():
                free_balance = Decimal(balance_entry["available"])
                total_balance = Decimal(balance_entry["available"]) + Decimal(balance_entry["locked"])
                self._account_available_balances[asset_name] = free_balance
                self._account_balances[asset_name] = total_balance
                remote_asset_names.add(asset_name)

            asset_names_to_remove = local_asset_names.difference(remote_asset_names)
            for asset_name in asset_names_to_remove:
                del self._account_available_balances[asset_name]
                del self._account_balances[asset_name]

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: List[Dict[str, Any]]):
        mapping = bidict()
        for symbol_data in exchange_info:
            if utils.is_exchange_information_valid(symbol_data):
                mapping[symbol_data["symbol"]] = combine_to_hb_trading_pair(base=symbol_data["baseSymbol"],
                                                                            quote=symbol_data["quoteSymbol"])
        self._set_trading_pair_symbol_map(mapping)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        params = {
            "symbol": self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        }

        resp_json = await self._api_request(
            method=RESTMethod.GET,
            path_url=CONSTANTS.TICKER_PRICE_CHANGE_PATH_URL,
            params=params
        )

        return float(resp_json["lastPrice"])

    @property
    def funding_fee_poll_interval(self) -> int:
        return 120

    def supported_position_modes(self) -> List:
        return [PositionMode.ONEWAY]

    def get_buy_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules.get(trading_pair)
        return trading_rule.buy_order_collateral_token

    def get_sell_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules.get(trading_pair)
        return trading_rule.sell_order_collateral_token

    async def _initialize_leverage_if_needed(self):
        """Fetch and initialize leverage from exchange if not already set."""
        if not self._leverage_initialized:
            try:
                account_info = await self._api_get(
                    path_url=CONSTANTS.ACCOUNT_PATH_URL,
                    is_auth_required=True
                )
                self._leverage = Decimal(str(account_info.get("leverageLimit", "1")))
                self._leverage_initialized = True
            except Exception as e:
                self.logger().warning(f"Failed to fetch leverage. Positions will be loaded on next polling loop: {e}")
                raise

    async def _update_positions(self):
        try:
            await self._initialize_leverage_if_needed()
        except Exception:
            return

        params = {
            "instruction": "positionQuery",
        }
        try:
            positions = await self._api_get(path_url=CONSTANTS.POSITIONS_PATH_URL,
                                            params=params,
                                            is_auth_required=True)
            for position in positions:
                trading_pair = position.get("symbol")
                try:
                    hb_trading_pair = self.trading_pair_associated_to_exchange_symbol(trading_pair)
                except KeyError:
                    # Ignore results for which their symbols is not tracked by the connector
                    continue
                unrealized_pnl = Decimal(position.get("pnlUnrealized"))
                entry_price = Decimal(position.get("entryPrice"))
                net_quantity = Decimal(position.get("netQuantity", "0"))
                amount = abs(net_quantity)
                position_side = PositionSide.SHORT if net_quantity < 0 else PositionSide.LONG
                pos_key = self._perpetual_trading.position_key(hb_trading_pair, position_side)
                if amount != 0:
                    _position = Position(
                        trading_pair=self.trading_pair_associated_to_exchange_symbol(trading_pair),
                        position_side=position_side,
                        unrealized_pnl=unrealized_pnl,
                        entry_price=entry_price,
                        amount=amount,
                        leverage=self._leverage
                    )
                    self._perpetual_trading.set_position(pos_key, _position)
                else:
                    self._perpetual_trading.remove_position(pos_key)
        except Exception as e:
            self.logger().error(f"Error fetching positions: {e}", exc_info=True)

    async def _trading_pair_position_mode_set(self, mode: PositionMode, trading_pair: str) -> Tuple[bool, str]:
        """
        :return: A tuple of boolean (true if success) and error message if the exchange returns one on failure.
        """
        if mode != PositionMode.ONEWAY:
            self.trigger_event(
                AccountEvent.PositionModeChangeFailed,
                PositionModeChangeEvent(
                    self.current_timestamp, trading_pair, mode, "Backpack only supports the ONEWAY position mode."
                ),
            )
            self.logger().debug(
                f"Backpack encountered a problem switching position mode to "
                f"{mode} for {trading_pair}"
                f" (Backpack only supports the ONEWAY position mode)"
            )
        else:
            self._position_mode = PositionMode.ONEWAY
            super().set_position_mode(PositionMode.ONEWAY)
            self.trigger_event(
                AccountEvent.PositionModeChangeSucceeded,
                PositionModeChangeEvent(self.current_timestamp, trading_pair, mode),
            )
            self.logger().debug(f"Backpack switching position mode to " f"{mode} for {trading_pair} succeeded.")
        return True, ""

    async def _set_trading_pair_leverage(self, trading_pair: str, leverage: int) -> Tuple[bool, str]:
        if not leverage:
            return False, f"There is no leverage available for {trading_pair}."

        data = {
            "instruction": "accountUpdate",
            "leverageLimit": str(leverage),
        }
        try:
            # Backpack returns 200 with no content
            rest_assistant = await self._web_assistants_factory.get_rest_assistant()
            url = web_utils.private_rest_url(path_url=CONSTANTS.ACCOUNT_PATH_URL, domain=self._domain)

            response = await rest_assistant.execute_request_and_get_response(
                url=url,
                data=data,
                method=RESTMethod.PATCH,
                is_auth_required=True,
                throttler_limit_id=CONSTANTS.ACCOUNT_PATH_URL,
            )

            # Check if status is 2xx (success)
            if 200 <= response.status < 300:
                self.logger().info(f"Successfully set leverage to {leverage} for account")
                self._leverage = Decimal(str(leverage))
                self._leverage_initialized = True
                return True, ""
            else:
                error_text = await response.text()
                error_msg = f"Failed to set leverage: HTTP {response.status} - {error_text}"
                self.logger().error(error_msg)
                return False, error_msg
        except Exception as e:
            error_msg = f"Error setting leverage for {trading_pair}: {str(e)}"
            self.logger().error(error_msg, exc_info=True)
            return False, error_msg

    async def _fetch_last_fee_payment(self, trading_pair: str) -> Tuple[float, Decimal, Decimal]:
        params = {
            "instruction": "fundingHistoryQueryAll",
            "symbol": self.exchange_symbol_associated_to_pair(trading_pair=trading_pair),
            "sortDirection": "Desc",
        }
        funding_payment_info = await self._api_get(path_url=CONSTANTS.FUNDING_PAYMENTS_PATH_URL,
                                                   params=params,
                                                   is_auth_required=True)
        if not funding_payment_info:
            return 0, Decimal("-1"), Decimal("-1")
        last_payment = funding_payment_info[0]
        if last_payment:
            timestamp = pd.Timestamp(last_payment["intervalEndTimestamp"]).timestamp()
            rate = Decimal(last_payment["fundingRate"])
            amount = Decimal(last_payment["quantity"])
            return timestamp, rate, amount
        return 0, Decimal("-1"), Decimal("-1")



from decimal import Decimal
from typing import TYPE_CHECKING, Dict, Optional

from hummingbot.connector.utils import split_hb_trading_pair
from hummingbot.core.rate_oracle.sources.rate_source_base import RateSourceBase
from hummingbot.core.utils import async_ttl_cache

if TYPE_CHECKING:
    from hummingbot.connector.derivative.pacifica_perpetual.pacifica_perpetual_derivative import (
        PacificaPerpetualDerivative,
    )


class PacificaPerpetualRateSource(RateSourceBase):
    def __init__(self):
        super().__init__()
        self._exchange: Optional[PacificaPerpetualDerivative] = None

    @property
    def name(self) -> str:
        return "pacifica_perpetual"

    @async_ttl_cache(ttl=30, maxsize=1)
    async def get_prices(self, quote_token: Optional[str] = None) -> Dict[str, Decimal]:
        if quote_token is not None and quote_token != "USDC":
            raise ValueError("Pacifica Perpetual only supports USDC as quote token.")

        self._ensure_exchange()
        results = {}
        try:
            pairs_prices = await self._exchange.get_all_pairs_prices()
            for pair_price in pairs_prices:
                trading_pair = pair_price["trading_pair"]
                if quote_token is not None:
                    base, quote = split_hb_trading_pair(trading_pair=trading_pair)
                    if quote != quote_token:
                        continue
                price = pair_price["price"]
                results[trading_pair] = Decimal(price)
        except Exception:
            self.logger().exception(
                msg="Unexpected error while retrieving rates from Pacifica. Check the log file for more info.",
            )
        return results

    def _ensure_exchange(self):
        if self._exchange is None:
            self._exchange = self._build_pacifica_connector_without_private_keys()

    @staticmethod
    def _build_pacifica_connector_without_private_keys() -> 'PacificaPerpetualDerivative':
        from hummingbot.connector.derivative.pacifica_perpetual.pacifica_perpetual_derivative import (
            PacificaPerpetualDerivative,
        )

        return PacificaPerpetualDerivative(
            pacifica_perpetual_agent_wallet_public_key="dummy_public_key",
            pacifica_perpetual_agent_wallet_private_key="5Bnr1LtPzXwFBd8z4F1ceR42yycUeUt4zoiDChW7cLDzVD6SmbHwwFhwbbLDExscxeVBbW6WVbWTKX4Dse4WUung",  # dummy 64-byte base58 encoded key
            pacifica_perpetual_user_wallet_public_key="dummy_user_key",
            trading_pairs=[],
            trading_required=False,
        )



import asyncio
from typing import TYPE_CHECKING, List, Optional

from hummingbot.connector.exchange.backpack import backpack_constants as CONSTANTS
from hummingbot.connector.exchange.backpack.backpack_auth import BackpackAuth
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.backpack.backpack_exchange import BackpackExchange


class BackpackAPIUserStreamDataSource(UserStreamTrackerDataSource):

    LISTEN_KEY_KEEP_ALIVE_INTERVAL = 60  # Recommended to Ping/Update listen key to keep connection alive
    HEARTBEAT_TIME_INTERVAL = 30.0
    LISTEN_KEY_RETRY_INTERVAL = 5.0
    MAX_RETRIES = 3

    _logger: Optional[HummingbotLogger] = None

    def __init__(self,
                 auth: AuthBase,
                 trading_pairs: List[str],
                 connector: 'BackpackExchange',
                 api_factory: WebAssistantsFactory,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN):
        super().__init__()
        self._auth: BackpackAuth = auth
        self._domain = domain
        self._api_factory = api_factory
        self._connector = connector

    async def _get_ws_assistant(self) -> WSAssistant:
        """
        Creates a new WSAssistant instance.
        """
        # Always create a new assistant to avoid connection issues
        return await self._api_factory.get_ws_assistant()

    async def _connected_websocket_assistant(self) -> WSAssistant:
        """
        Creates an instance of WSAssistant connected to the exchange.

        This method ensures the listen key is ready before connecting.
        """
        # Get a websocket assistant and connect it
        ws = await self._get_ws_assistant()
        url = f"{CONSTANTS.WSS_URL.format(self._domain)}"

        await ws.connect(ws_url=url, ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        self.logger().info("Successfully connected to user stream")

        return ws

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.

        :param websocket_assistant: the websocket assistant used to connect to the exchange
        """
        try:
            timestamp_ms = int(self._auth.time_provider.time() * 1e3)
            signature = self._auth.generate_signature(params={},
                                                      timestamp_ms=timestamp_ms,
                                                      window_ms=self._auth.DEFAULT_WINDOW_MS,
                                                      instruction="subscribe")
            orders_change_payload = {
                "method": "SUBSCRIBE",
                "params": [CONSTANTS.ALL_ORDERS_CHANNEL],
                "signature": [
                    self._auth.api_key,
                    signature,
                    str(timestamp_ms),
                    str(self._auth.DEFAULT_WINDOW_MS)
                ]
            }
            subscribe_order_change_request: WSJSONRequest = WSJSONRequest(payload=orders_change_payload)

            await websocket_assistant.send(subscribe_order_change_request)

            self.logger().info("Subscribed to private order changes channel...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to user streams...")
            raise

    async def _on_user_stream_interruption(self, websocket_assistant: Optional[WSAssistant]):
        """
        Handles websocket disconnection by cleaning up resources.

        :param websocket_assistant: The websocket assistant that was disconnected
        """
        # Disconnect the websocket if it exists
        websocket_assistant and await websocket_assistant.disconnect()



import asyncio
from typing import TYPE_CHECKING, Any, Dict, List, NoReturn, Optional

from hummingbot.connector.exchange.bitget import bitget_constants as CONSTANTS, bitget_web_utils as web_utils
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest, WSPlainTextRequest
from hummingbot.core.web_assistant.rest_assistant import RESTAssistant
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant

if TYPE_CHECKING:
    from hummingbot.connector.exchange.bitget.bitget_exchange import BitgetExchange


class BitgetAPIOrderBookDataSource(OrderBookTrackerDataSource):
    """
    Data source for retrieving order book data from the Bitget exchange via REST and WebSocket APIs.
    """
    _DYNAMIC_SUBSCRIBE_ID_START = 100
    _next_subscribe_id: int = _DYNAMIC_SUBSCRIBE_ID_START

    def __init__(
        self,
        trading_pairs: List[str],
        connector: 'BitgetExchange',
        api_factory: WebAssistantsFactory,
    ) -> None:
        super().__init__(trading_pairs)
        self._connector: 'BitgetExchange' = connector
        self._api_factory: WebAssistantsFactory = api_factory
        self._ping_task: Optional[asyncio.Task] = None

    async def get_last_traded_prices(
        self,
        trading_pairs: List[str],
        domain: Optional[str] = None
    ) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def _parse_pong_message(self) -> None:
        self.logger().debug("PING-PONG message for order book completed")

    async def _process_message_for_unknown_channel(
        self,
        event_message: Dict[str, Any],
        websocket_assistant: WSAssistant,
    ) -> None:
        if event_message == CONSTANTS.PUBLIC_WS_PONG_RESPONSE:
            await self._parse_pong_message()
        elif "event" in event_message:
            if event_message["event"] == "error":
                message = event_message.get("msg", "Unknown error")
                error_code = event_message.get("code", "Unknown code")
                raise IOError(f"Failed to subscribe to public channels: {message} ({error_code})")

            if event_message["event"] == "subscribe":
                channel: str = event_message["arg"]["channel"]
                self.logger().info(f"Subscribed to public channel: {channel.upper()}")
        else:
            self.logger().info(f"Message for unknown channel received: {event_message}")

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> Optional[str]:
        channel: Optional[str] = None

        if "arg" in event_message and "action" in event_message:
            arg: Dict[str, Any] = event_message["arg"]
            response_channel: Optional[str] = arg.get("channel")

            if response_channel == CONSTANTS.PUBLIC_WS_BOOKS:
                action: Optional[str] = event_message.get("action")
                channels = {
                    "snapshot": self._snapshot_messages_queue_key,
                    "update": self._diff_messages_queue_key
                }
                channel = channels.get(action)
            elif response_channel == CONSTANTS.PUBLIC_WS_TRADE:
                channel = self._trade_messages_queue_key

        return channel

    async def _parse_any_order_book_message(
        self,
        data: Dict[str, Any],
        symbol: str,
        message_type: OrderBookMessageType,
    ) -> OrderBookMessage:
        """
        Parse a WebSocket message into an OrderBookMessage for snapshots or diffs.

        :param raw_message: The raw WebSocket message.
        :param message_type: The type of order book message (SNAPSHOT or DIFF).

        :return: The parsed order book message.
        """
        trading_pair: str = await self._connector.trading_pair_associated_to_exchange_symbol(symbol)
        update_id: int = int(data["ts"])
        timestamp: float = update_id * 1e-3

        order_book_message_content: Dict[str, Any] = {
            "trading_pair": trading_pair,
            "update_id": update_id,
            "bids": data["bids"],
            "asks": data["asks"],
        }

        return OrderBookMessage(
            message_type=message_type,
            content=order_book_message_content,
            timestamp=timestamp
        )

    async def _parse_order_book_diff_message(
        self,
        raw_message: Dict[str, Any],
        message_queue: asyncio.Queue
    ) -> None:
        diffs_data: Dict[str, Any] = raw_message["data"]
        symbol: str = raw_message["arg"]["instId"]

        for diff in diffs_data:
            diff_message: OrderBookMessage = await self._parse_any_order_book_message(
                data=diff,
                symbol=symbol,
                message_type=OrderBookMessageType.DIFF
            )

            message_queue.put_nowait(diff_message)

    async def _parse_order_book_snapshot_message(
        self,
        raw_message: Dict[str, Any],
        message_queue: asyncio.Queue
    ) -> None:
        snapshot_data: Dict[str, Any] = raw_message["data"]
        symbol: str = raw_message["arg"]["instId"]

        for snapshot in snapshot_data:
            snapshot_message: OrderBookMessage = await self._parse_any_order_book_message(
                data=snapshot,
                symbol=symbol,
                message_type=OrderBookMessageType.SNAPSHOT
            )

            message_queue.put_nowait(snapshot_message)

    async def _parse_trade_message(
        self,
        raw_message: Dict[str, Any],
        message_queue: asyncio.Queue
    ) -> None:
        data: List[Dict[str, Any]] = raw_message["data"]
        symbol: str = raw_message["arg"]["instId"]
        trading_pair: str = await self._connector.trading_pair_associated_to_exchange_symbol(symbol)

        for trade_data in data:
            trade_type: float = float(TradeType.BUY.value) \
                if trade_data["side"] == "buy" else float(TradeType.SELL.value)
            message_content: Dict[str, Any] = {
                "trade_id": int(trade_data["tradeId"]),
                "trading_pair": trading_pair,
                "trade_type": trade_type,
                "amount": trade_data["size"],
                "price": trade_data["price"],
            }
            trade_message = OrderBookMessage(
                message_type=OrderBookMessageType.TRADE,
                content=message_content,
                timestamp=int(trade_data["ts"]) * 1e-3,
            )
            message_queue.put_nowait(trade_message)

    async def _connected_websocket_assistant(self) -> WSAssistant:
        websocket_assistant: WSAssistant = await self._api_factory.get_ws_assistant()

        await websocket_assistant.connect(
            ws_url=web_utils.public_ws_url(),
            message_timeout=CONSTANTS.SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE,
        )

        return websocket_assistant

    async def _subscribe_channels(self, ws: WSAssistant) -> None:
        try:
            subscription_topics: List[Dict[str, str]] = []

            for trading_pair in self._trading_pairs:
                symbol: str = await self._connector.exchange_symbol_associated_to_pair(
                    trading_pair
                )
                for channel in [CONSTANTS.PUBLIC_WS_BOOKS, CONSTANTS.PUBLIC_WS_TRADE]:
                    subscription_topics.append({
                        "instType": "SPOT",
                        "channel": channel,
                        "instId": symbol
                    })

            await ws.send(
                WSJSONRequest({
                    "op": "subscribe",
                    "args": subscription_topics,
                })
            )

            self.logger().info("Subscribed to public channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to public channels...")
            raise

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        symbol: str = await self._connector.exchange_symbol_associated_to_pair(trading_pair)
        rest_assistant: RESTAssistant = await self._api_factory.get_rest_assistant()

        data: Dict[str, Any] = await rest_assistant.execute_request(
            url=web_utils.public_rest_url(path_url=CONSTANTS.PUBLIC_ORDERBOOK_ENDPOINT),
            params={
                "symbol": symbol,
                "limit": "100",
            },
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.PUBLIC_ORDERBOOK_ENDPOINT,
        )

        return data

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot_response: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_data: Dict[str, Any] = snapshot_response["data"]
        update_id: int = int(snapshot_data["ts"])
        timestamp: float = update_id * 1e-3

        order_book_message_content: Dict[str, Any] = {
            "trading_pair": trading_pair,
            "update_id": update_id,
            "bids": snapshot_data["bids"],
            "asks": snapshot_data["asks"],
        }

        return OrderBookMessage(
            OrderBookMessageType.SNAPSHOT,
            order_book_message_content,
            timestamp
        )

    async def _send_ping(self, websocket_assistant: WSAssistant) -> None:
        ping_request = WSPlainTextRequest(CONSTANTS.PUBLIC_WS_PING_REQUEST)

        await websocket_assistant.send(ping_request)

    async def send_interval_ping(self, websocket_assistant: WSAssistant) -> None:
        """
        Coroutine to send PING messages periodically.

        :param websocket_assistant: The websocket assistant to use to send the PING message.
        """
        try:
            while True:
                await self._send_ping(websocket_assistant)
                await asyncio.sleep(CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        except asyncio.CancelledError:
            self.logger().info("Interval PING task cancelled")
            raise
        except Exception:
            self.logger().exception("Error sending interval PING")

    async def listen_for_subscriptions(self) -> NoReturn:
        ws: Optional[WSAssistant] = None
        while True:
            try:
                ws: WSAssistant = await self._connected_websocket_assistant()
                self._ws_assistant = ws  # Store for dynamic subscriptions
                await self._subscribe_channels(ws)
                self._ping_task = asyncio.create_task(self.send_interval_ping(ws))
                await self._process_websocket_messages(websocket_assistant=ws)
            except asyncio.CancelledError:
                raise
            except ConnectionError as connection_exception:
                self.logger().warning(
                    f"The websocket connection was closed ({connection_exception})"
                )
            except Exception:
                self.logger().exception(
                    "Unexpected error occurred when listening to order book streams. "
                    "Retrying in 5 seconds...",
                )
                await self._sleep(1.0)
            finally:
                self._ws_assistant = None
                if self._ping_task is not None:
                    self._ping_task.cancel()
                    try:
                        await self._ping_task
                    except asyncio.CancelledError:
                        pass
                    self._ping_task = None
                await self._on_order_stream_interruption(websocket_assistant=ws)

    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        """
        Subscribes to order book and trade channels for a single trading pair
        on the existing WebSocket connection.

        :param trading_pair: the trading pair to subscribe to
        :return: True if subscription was successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot subscribe to {trading_pair}: WebSocket not connected"
            )
            return False

        try:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair)

            subscription_topics = []
            for channel in [CONSTANTS.PUBLIC_WS_BOOKS, CONSTANTS.PUBLIC_WS_TRADE]:
                subscription_topics.append({
                    "instType": "SPOT",
                    "channel": channel,
                    "instId": symbol
                })

            await self._ws_assistant.send(
                WSJSONRequest({
                    "op": "subscribe",
                    "args": subscription_topics,
                })
            )

            self.add_trading_pair(trading_pair)
            self.logger().info(f"Subscribed to {trading_pair} order book and trade channels")
            return True

        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception(f"Error subscribing to {trading_pair}")
            return False

    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        """
        Unsubscribes from order book and trade channels for a single trading pair
        on the existing WebSocket connection.

        :param trading_pair: the trading pair to unsubscribe from
        :return: True if unsubscription was successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot unsubscribe from {trading_pair}: WebSocket not connected"
            )
            return False

        try:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair)

            unsubscription_topics = []
            for channel in [CONSTANTS.PUBLIC_WS_BOOKS, CONSTANTS.PUBLIC_WS_TRADE]:
                unsubscription_topics.append({
                    "instType": "SPOT",
                    "channel": channel,
                    "instId": symbol
                })

            await self._ws_assistant.send(
                WSJSONRequest({
                    "op": "unsubscribe",
                    "args": unsubscription_topics,
                })
            )

            self.remove_trading_pair(trading_pair)
            self.logger().info(f"Unsubscribed from {trading_pair} order book and trade channels")
            return True

        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception(f"Error unsubscribing from {trading_pair}")
            return False

    @classmethod
    def _get_next_subscribe_id(cls) -> int:
        """Returns the next subscription ID and increments the counter."""
        current_id = cls._next_subscribe_id
        cls._next_subscribe_id += 1
        return current_id



import asyncio
from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple

from bidict import bidict

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.exchange.ascend_ex import (
    ascend_ex_constants as CONSTANTS,
    ascend_ex_utils as utils,
    ascend_ex_web_utils as web_utils,
)
from hummingbot.connector.exchange.ascend_ex.ascend_ex_api_order_book_data_source import AscendExAPIOrderBookDataSource
from hummingbot.connector.exchange.ascend_ex.ascend_ex_api_user_stream_data_source import (
    AscendExAPIUserStreamDataSource,
)
from hummingbot.connector.exchange.ascend_ex.ascend_ex_auth import AscendExAuth
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair, split_hb_trading_pair
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import AddedToCostTradeFee, TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class AscendExExchange(ExchangePyBase):
    """
    AscendExExchange connects with AscendEx exchange and provides order book pricing, user account tracking and
    trading functionality.
    """

    UPDATE_ORDER_STATUS_MIN_INTERVAL = 10.0

    web_utils = web_utils

    def __init__(
        self,
        ascend_ex_api_key: str,
        ascend_ex_secret_key: str,
        ascend_ex_group_id: str,
        balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
        rate_limits_share_pct: Decimal = Decimal("100"),
        trading_pairs: Optional[List[str]] = None,
        trading_required: bool = True,
    ):
        """
        :param client_config_map: The config map of the client instance.
        :param ascend_ex_api_key: The API key to connect to private AscendEx APIs.
        :param ascend_ex_secret_key: The API secret.
        :param trading_pairs: The market trading pairs which to track order book data.
        :param trading_required: Whether actual trading is needed.
        """
        self.ascend_ex_api_key = ascend_ex_api_key
        self.ascend_ex_secret_key = ascend_ex_secret_key
        self.ascend_ex_group_id = ascend_ex_group_id
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        super().__init__(balance_asset_limit, rate_limits_share_pct)

        self._last_known_sequence_number = 0

    @property
    def domain(self):
        return CONSTANTS.DEFAULT_DOMAIN

    @property
    def authenticator(self):
        return AscendExAuth(self.ascend_ex_api_key, self.ascend_ex_secret_key)

    @property
    def name(self) -> str:
        return CONSTANTS.EXCHANGE_NAME

    @property
    def rate_limits_rules(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def client_order_id_max_length(self):
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self):
        return CONSTANTS.HBOT_ORDER_ID_PREFIX

    @property
    def trading_rules_request_path(self):
        return CONSTANTS.PRODUCTS_PATH_URL

    @property
    def trading_pairs_request_path(self):
        return CONSTANTS.PRODUCTS_PATH_URL

    @property
    def check_network_request_path(self):
        return CONSTANTS.SERVER_LIMIT_INFO

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return False

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    def supported_order_types(self):
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.MARKET]

    async def get_all_pairs_prices(self) -> Dict[str, Any]:
        """
        This method executes a request to the exchange to get the current price for all trades.
        It returns the response of the exchange (expected to be used by the AscendEx RateSource for the RateOracle)

        :return: the response from the tickers endpoint
        """
        symbol_to_trading_pair_map = await self.trading_pair_symbol_map()
        pairs_prices = await self._api_get(path_url=CONSTANTS.TICKER_PATH_URL)
        spot_valid_token_entries = [
            data_dict for data_dict in pairs_prices["data"] if data_dict["symbol"] in symbol_to_trading_pair_map
        ]
        pairs_prices["data"] = spot_valid_token_entries
        return pairs_prices

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        # API documentation does not clarify the error message for timestamp related problems
        return False

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_lost_order_removed_if_not_found_during_order_status_update
        # when replacing the dummy implementation
        return False

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_cancel_order_not_found_in_the_exchange when replacing the
        # dummy implementation
        return False

    async def _api_request_url(self, path_url: str, is_auth_required: bool = False) -> str:
        url = await super()._api_request_url(path_url, is_auth_required)

        if is_auth_required:
            url = url.format(group_id=self.ascend_ex_group_id)

        return url

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(throttler=self._throttler, auth=self._auth)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return AscendExAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return AscendExAPIUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
        )

    def _get_fee(
        self,
        base_currency: str,
        quote_currency: str,
        order_type: OrderType,
        order_side: TradeType,
        amount: Decimal,
        price: Decimal = s_decimal_NaN,
        is_maker: Optional[bool] = None,
    ) -> AddedToCostTradeFee:

        is_maker = is_maker or (order_type is OrderType.LIMIT_MAKER)
        trading_pair = combine_to_hb_trading_pair(base=base_currency, quote=quote_currency)
        if trading_pair in self._trading_fees:
            fees_data = self._trading_fees[trading_pair]
            fee_value = Decimal(fees_data["maker"]) if is_maker else Decimal(fees_data["taker"])
            fee = AddedToCostTradeFee(percent=fee_value)
        else:
            fee = build_trade_fee(
                self.name,
                is_maker,
                base_currency=base_currency,
                quote_currency=quote_currency,
                order_type=order_type,
                order_side=order_side,
                amount=amount,
                price=price,
            )
        return fee

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: Dict[str, Any]):
        mapping = bidict()
        for symbol_data in filter(utils.is_pair_information_valid, exchange_info.get("data", [])):
            if len(symbol_data["symbol"].split("/")) == 2:
                base, quote = symbol_data["symbol"].split("/")
                mapping[symbol_data["symbol"]] = combine_to_hb_trading_pair(base, quote)
        self._set_trading_pair_symbol_map(mapping)

    async def _place_order(
        self,
        order_id: str,
        trading_pair: str,
        amount: Decimal,
        trade_type: TradeType,
        order_type: OrderType,
        price: Decimal,
        **kwargs,
    ) -> Tuple[str, float]:
        side = trade_type.name.lower()
        timestamp = utils.get_ms_timestamp()
        data = {
            "time": timestamp,
            "orderQty": str(amount),
            "id": order_id,
            "side": side,
            "symbol": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair),
        }
        if order_type.is_limit_type():
            data["orderPrice"] = str(price)
            data["orderType"] = "limit"
            data["timeInForce"] = "GTC"
        else:
            data["orderType"] = "market"
            data["timeInForce"] = "IOC"
        if order_type is OrderType.LIMIT_MAKER:
            data["postOnly"] = True
        exchange_order = await self._api_post(
            path_url=CONSTANTS.ORDER_PATH_URL,
            data=data,
            is_auth_required=True,
        )

        if exchange_order.get("code") == 0:
            return (
                str(exchange_order["data"]["info"]["orderId"]),
                int(exchange_order["data"]["info"].get("timestamp") or exchange_order["data"]["info"]
                    ["lastExecTime"]) * 1e-3,
            )
        else:
            raise IOError(str(exchange_order))

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        """
        This implementation specific function is called by _cancel, and returns True if successful
        """
        exchange_order_id = await tracked_order.get_exchange_order_id()
        timestamp = utils.get_ms_timestamp()
        data = {
            "time": timestamp,
            "orderId": exchange_order_id,
            "symbol": await self.exchange_symbol_associated_to_pair(trading_pair=tracked_order.trading_pair),
        }
        cancel_result = await self._api_delete(
            path_url=CONSTANTS.ORDER_PATH_URL,
            data=data,
            is_auth_required=True,
        )
        if cancel_result.get("code") == 0:
            return True
        return False

    async def _user_stream_event_listener(self):
        """
        This functions runs in background continuously processing the events received from the exchange by the user
        stream data source. It keeps reading events from the queue until the task is interrupted.
        The events received are balance updates, order updates and trade events.
        """
        async for event_message in self._iter_user_event_queue():
            try:
                acct_type = event_message.get("ac")
                event_subject = event_message.get("m")
                execution_data = event_message.get("data")

                # Refer to https://ascendex.github.io/ascendex-pro-api/#channel-order-and-balance
                if acct_type == CONSTANTS.ACCOUNT_TYPE and event_subject == CONSTANTS.ORDER_CHANGE_EVENT_TYPE:
                    order_event_type = execution_data["st"]
                    order_id: Optional[str] = execution_data.get("orderId")
                    event_timestamp = execution_data["t"] * 1e-3
                    updated_status = CONSTANTS.ORDER_STATE[order_event_type]

                    fillable_order_list = list(
                        filter(
                            lambda order: order.exchange_order_id == order_id,
                            list(self._order_tracker.all_fillable_orders.values()),
                        )
                    )
                    updatable_order_list = list(
                        filter(
                            lambda order: order.exchange_order_id == order_id,
                            list(self._order_tracker.all_updatable_orders.values()),
                        )
                    )

                    fillable_order = None
                    if len(fillable_order_list) > 0:
                        fillable_order = fillable_order_list[0]

                    updatable_order = None
                    if len(updatable_order_list) > 0:
                        updatable_order = updatable_order_list[0]

                    if fillable_order is not None and updated_status in [
                        OrderState.PARTIALLY_FILLED,
                        OrderState.FILLED,
                    ]:
                        executed_amount_diff = Decimal(execution_data["cfq"]) - fillable_order.executed_amount_base
                        execute_price = Decimal(execution_data["ap"])
                        fee_asset = execution_data["fa"]
                        total_order_fee = Decimal(execution_data["cf"])
                        current_accumulated_fee = 0
                        for fill in fillable_order.order_fills.values():
                            current_accumulated_fee += sum(
                                (fee.amount for fee in fill.fee.flat_fees if fee.token == fee_asset)
                            )

                        fee = TradeFeeBase.new_spot_fee(
                            fee_schema=self.trade_fee_schema(),
                            trade_type=fillable_order.trade_type,
                            percent_token=fee_asset,
                            flat_fees=[TokenAmount(amount=total_order_fee - current_accumulated_fee, token=fee_asset)],
                        )

                        trade_update = TradeUpdate(
                            trade_id=str(execution_data["sn"]),
                            client_order_id=fillable_order.client_order_id,
                            exchange_order_id=order_id,
                            trading_pair=updatable_order.trading_pair,
                            fee=fee,
                            fill_base_amount=executed_amount_diff,
                            fill_quote_amount=executed_amount_diff * execute_price,
                            fill_price=execute_price,
                            fill_timestamp=event_timestamp,
                        )
                        self._order_tracker.process_trade_update(trade_update)

                    if updatable_order is not None:
                        order_update = OrderUpdate(
                            trading_pair=updatable_order.trading_pair,
                            update_timestamp=event_timestamp,
                            new_state=updated_status,
                            client_order_id=fillable_order.client_order_id,
                            exchange_order_id=order_id,
                        )
                        self._order_tracker.process_order_update(order_update=order_update)

                    # Update the balance with the balance status details included in the order event
                    trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=execution_data["s"])
                    base_asset, quote_asset = split_hb_trading_pair(trading_pair=trading_pair)
                    self._account_balances.update({base_asset: Decimal(execution_data["btb"])})
                    self._account_available_balances.update({base_asset: Decimal(execution_data["bab"])})
                    self._account_balances.update({quote_asset: Decimal(execution_data["qtb"])})
                    self._account_available_balances.update({quote_asset: Decimal(execution_data["qab"])})

                # The balance event is not processed because it only sends transfers information
                # We need to use the offline balance estimation for AscendEx

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error in user stream listener loop.")
                await self._sleep(5.0)

    async def _update_balances(self):
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        response = await self._api_get(path_url=CONSTANTS.BALANCE_PATH_URL, is_auth_required=True)

        if response.get("code") == 0:
            for balance_entry in response["data"]:
                asset_name = balance_entry["asset"]
                self._account_available_balances[asset_name] = Decimal(balance_entry["availableBalance"])
                self._account_balances[asset_name] = Decimal(balance_entry["totalBalance"])
                remote_asset_names.add(asset_name)

            asset_names_to_remove = local_asset_names.difference(remote_asset_names)
            for asset_name in asset_names_to_remove:
                del self._account_available_balances[asset_name]
                del self._account_balances[asset_name]
        else:
            self.logger().error(f"There was an error during the balance request to AscendEx ({response})")
            raise IOError(f"Error requesting balances from AscendEx ({response})")

    async def _format_trading_rules(self, raw_trading_pair_info: Dict[str, Any]) -> List[TradingRule]:
        trading_rules = []

        for info in filter(utils.is_pair_information_valid, raw_trading_pair_info.get("data", [])):
            try:
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=info.get("symbol"))
                trading_rules.append(
                    TradingRule(
                        trading_pair=trading_pair,
                        min_order_size=Decimal(info["minQty"]),
                        max_order_size=Decimal(info["maxQty"]),
                        min_price_increment=Decimal(info["tickSize"]),
                        min_base_amount_increment=Decimal(info["lotSize"]),
                        min_notional_size=Decimal(info["minNotional"]),
                    )
                )
            except Exception:
                self.logger().exception(f"Error parsing the trading pair rule {info}. Skipping.", exc_info=True)
        return trading_rules

    async def _update_trading_fees(self):
        resp = await self._api_get(
            path_url=CONSTANTS.FEE_PATH_URL,
            is_auth_required=True,
        )
        fees_json = resp.get("data", {}).get("fees", [])
        for fee_json in fees_json:
            try:
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=fee_json["symbol"])
                self._trading_fees[trading_pair] = fee_json["fee"]
            except asyncio.CancelledError:
                raise
            except Exception:
                pass

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        # AscendEx does not have an endpoint to retrieve trades for a particular order
        # Thus it overrides the _update_orders_fills method
        pass

    def _trade_update_from_fill_data(self, fill_data: Dict[str, Any], order: InFlightOrder) -> TradeUpdate:
        trade_id = str(fill_data["sn"])
        timestamp = fill_data["transactTime"] * 1e3
        asset_amount_detail = {}
        fee_amount = 0
        fee_asset = order.quote_asset

        for asset_detail in fill_data["data"]:
            asset = asset_detail["asset"]
            amount = abs(Decimal(str(asset_detail["deltaQty"])))
            if asset_detail["dataType"] == "fee":
                fee_asset = asset
                fee_amount = amount
            else:
                asset_amount_detail[asset] = amount

        fee = TradeFeeBase.new_spot_fee(
            fee_schema=self.trade_fee_schema(),
            trade_type=order.trade_type,
            percent_token=fee_asset,
            flat_fees=[TokenAmount(amount=fee_amount, token=fee_asset)],
        )
        trade_update = TradeUpdate(
            trade_id=trade_id,
            client_order_id=order.client_order_id,
            exchange_order_id=order.exchange_order_id,
            trading_pair=order.trading_pair,
            fee=fee,
            fill_base_amount=asset_amount_detail[order.base_asset],
            fill_quote_amount=asset_amount_detail[order.quote_asset],
            fill_price=asset_amount_detail[order.quote_asset] / asset_amount_detail[order.base_asset],
            fill_timestamp=timestamp,
        )

        return trade_update

    async def _all_trade_updates_for_orders(
        self, orders: List[InFlightOrder], sequence_number: int
    ) -> Tuple[List[TradeUpdate], int]:
        # This endpoint determines the URL in an adhoc way because it is very different compare to the other endpoints
        url = await self._api_request_url(path_url="")
        balance_hist_url = url.replace("/v1/", f"/{CONSTANTS.BALANCE_HISTORY_PATH_URL}")
        params = {"sn": sequence_number, "limit": 500}
        trade_updates = []
        orders_to_process = {order.exchange_order_id: order for order in orders if order.exchange_order_id is not None}
        should_request_next_page = True
        max_sequence_number = -1

        # If there are many pages of result, query at most two pages each time, to not delay the update status loop
        for _ in range(2):
            result = await self._api_get(
                path_url=CONSTANTS.BALANCE_HISTORY_PATH_URL,
                params=params,
                is_auth_required=True,
                overwrite_url=balance_hist_url,
            )

            if "order" in result:
                for order_fill_data in result["order"]:
                    max_sequence_number = max(max_sequence_number, order_fill_data["sn"])
                    if order_fill_data["orderId"] in orders_to_process:
                        order_id = order_fill_data["orderId"]
                        try:
                            trade_update = self._trade_update_from_fill_data(
                                fill_data=order_fill_data, order=orders_to_process[order_id]
                            )
                            trade_updates.append(trade_update)
                        except asyncio.CancelledError:
                            raise
                        except Exception as request_error:
                            self.logger().warning(
                                f"Failed to fetch trade updates for order {order_id}. Error: {request_error}"
                            )
                params["sn"] = max_sequence_number
                should_request_next_page = len(result["order"]) + len(result.get("balance", [])) == params["limit"]
                if not should_request_next_page:
                    break
            else:
                self.logger().warning(f"An error occurred when requesting order fills ({result})")
                break

        return trade_updates, max_sequence_number

    async def _update_orders_fills(self, orders: List[InFlightOrder]):
        if orders:
            # Since we are keeping the last order fill sequence number referenced to improve the query performance
            # it is necessary to evaluate updates for all possible fillable orders every time (to avoid loosing updates)
            candidate_orders = list(self._order_tracker.all_fillable_orders.values())
            try:
                if candidate_orders:
                    trade_updates, max_sequence_number = await self._all_trade_updates_for_orders(
                        orders=candidate_orders, sequence_number=self._last_known_sequence_number
                    )
                    # Update the _last_known_sequence_number to reduce the amount of information requested next time
                    self._last_known_sequence_number = max(self._last_known_sequence_number, max_sequence_number)
                    for trade_update in trade_updates:
                        self._order_tracker.process_trade_update(trade_update)
            except asyncio.CancelledError:
                raise
            except Exception as request_error:
                order_ids = [order.client_order_id for order in candidate_orders]
                self.logger().warning(f"Failed to fetch trade updates for orders {order_ids}. Error: {request_error}")

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        exchange_order_id = await tracked_order.get_exchange_order_id()
        params = {"orderId": exchange_order_id}
        updated_order_data = await self._api_get(
            path_url=CONSTANTS.ORDER_STATUS_PATH_URL, params=params, is_auth_required=True
        )

        if updated_order_data.get("code") == 0:
            order_update_data = updated_order_data["data"]
            ordered_state = order_update_data["status"]
            new_state = CONSTANTS.ORDER_STATE[ordered_state]

            order_update = OrderUpdate(
                client_order_id=tracked_order.client_order_id,
                exchange_order_id=order_update_data["orderId"],
                trading_pair=tracked_order.trading_pair,
                update_timestamp=self.current_timestamp,
                new_state=new_state,
            )

            return order_update
        else:
            raise IOError(f"Error requesting status for order {tracked_order.client_order_id} ({updated_order_data})")

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        params = {"symbol": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)}

        resp_json = await self._api_request(path_url=CONSTANTS.TICKER_PATH_URL, method=RESTMethod.GET, params=params)

        return float(resp_json["data"]["close"])



import decimal
import logging

import pandas as pd

FLOAT_PRINTOUT_PRECISION = 8


def format_decimal(n):
    """
    Convert the given float to a string without scientific notation
    """
    try:
        with decimal.localcontext() as ctx:
            if isinstance(n, float):
                n = ctx.create_decimal(n)
            if isinstance(n, decimal.Decimal):
                n = round(n, FLOAT_PRINTOUT_PRECISION)
                return format(n.normalize(), 'f')
            else:
                return str(n)
    except Exception as e:
        logging.getLogger().error(str(e))


pd.options.display.float_format = lambda x: format_decimal(x)



import argparse
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


class HelpCommand:
    def help(self,  # type: HummingbotApplication
             command: str):
        cmd_split = command.split()
        if cmd_split[0] == 'all':
            self.notify(self.parser.format_help())
        else:
            parser = self.parser._actions
            for step in cmd_split:
                subparsers_actions = [
                    action for action in parser if isinstance(action, argparse._SubParsersAction)]
                for subparsers_action in subparsers_actions:
                    subparser = subparsers_action.choices.get(step)
                    if subparser:
                        last_subparser = subparser
                        parser = subparser._actions
            self.notify(last_subparser.format_help())



# distutils: language=c++

from hummingbot.core.data_type.limit_order cimport LimitOrder
from hummingbot.core.time_iterator cimport TimeIterator


cdef class OrderTracker(TimeIterator):
    cdef:
        dict _tracked_limit_orders
        dict _tracked_market_orders
        dict _order_id_to_market_pair
        dict _shadow_tracked_limit_orders
        dict _shadow_order_id_to_market_pair
        object _shadow_gc_requests
        object _in_flight_cancels
        object _in_flight_pending_created

    cdef dict c_get_limit_orders(self)
    cdef dict c_get_market_orders(self)
    cdef dict c_get_shadow_limit_orders(self)
    cdef bint c_has_in_flight_cancel(self, str order_id)
    cdef bint c_check_and_track_cancel(self, str order_id)
    cdef object c_get_market_pair_from_order_id(self, str order_id)
    cdef object c_get_shadow_market_pair_from_order_id(self, str order_id)
    cdef LimitOrder c_get_limit_order(self, object market_pair, str order_id)
    cdef object c_get_market_order(self, object market_pair, str order_id)
    cdef LimitOrder c_get_shadow_limit_order(self, str order_id)
    cdef c_start_tracking_limit_order(self, object market_pair, str order_id, bint is_buy, object price,
                                      object quantity)
    cdef c_stop_tracking_limit_order(self, object market_pair, str order_id)
    cdef c_start_tracking_market_order(self, object market_pair, str order_id, bint is_buy, object quantity)
    cdef c_stop_tracking_market_order(self, object market_pair, str order_id)
    cdef c_check_and_cleanup_shadow_records(self)
    cdef c_add_create_order_pending(self, str order_id)
    cdef c_remove_create_order_pending(self, str order_id)



from typing import Any, Callable, Dict, List, Optional

from hummingbot.connector.derivative.okx_perpetual import okx_perpetual_constants as CONSTANTS
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.connector.utils import TimeSynchronizerRESTPreProcessor
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest
from hummingbot.core.web_assistant.rest_pre_processors import RESTPreProcessorBase
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class HeadersContentRESTPreProcessor(RESTPreProcessorBase):
    async def pre_process(self, request: RESTRequest) -> RESTRequest:
        request.headers = request.headers or {}
        request.headers.update({"Content-Type": "application/json"})
        return request


def build_api_factory(
        throttler: Optional[AsyncThrottler] = None,
        time_synchronizer: Optional[TimeSynchronizer] = None,
        time_provider: Optional[Callable] = None,
        auth: Optional[AuthBase] = None,
) -> WebAssistantsFactory:
    throttler = throttler or create_throttler()
    time_synchronizer = time_synchronizer or TimeSynchronizer()
    time_provider = time_provider or (lambda: get_current_server_time(throttler=throttler, domain=CONSTANTS.DEFAULT_DOMAIN))
    api_factory = WebAssistantsFactory(
        throttler=throttler,
        auth=auth,
        rest_pre_processors=[
            TimeSynchronizerRESTPreProcessor(synchronizer=time_synchronizer, time_provider=time_provider),
            HeadersContentRESTPreProcessor(),
        ],
    )
    return api_factory


def create_throttler(trading_pairs: List[str] = None) -> AsyncThrottler:
    throttler = AsyncThrottler(build_rate_limits(trading_pairs))
    return throttler


async def get_current_server_time(throttler: Optional[AsyncThrottler] = None,
                                  domain: str = CONSTANTS.DEFAULT_DOMAIN) -> float:
    """
    Transaction Timeouts (https://www.okx.com/docs-v5/en/?shell#overview-general-info)
    Orders may not be processed in time due to network delay or busy OKX servers.

    You can configure the expiry time of the request using expTime if you want the order request to be
    discarded after a specific time.

    If expTime is specified in the requests for Place (multiple) orders or Amend (multiple) orders,
    the request will not be processed if the current system time of the server is after the expTime.

    You should synchronize with our system time. Use Get system time to obtain the current system time.
    """
    throttler = throttler or create_throttler()
    api_factory = build_api_factory_without_time_synchronizer_pre_processor(throttler=throttler)
    rest_assistant = await api_factory.get_rest_assistant()
    endpoint = CONSTANTS.REST_SERVER_TIME[CONSTANTS.ENDPOINT]
    url = get_rest_url_for_endpoint(endpoint=endpoint, domain=domain)
    limit_id = get_rest_api_limit_id_for_endpoint(method=CONSTANTS.REST_SERVER_TIME[CONSTANTS.METHOD],
                                                  endpoint=endpoint)
    response = await rest_assistant.execute_request(
        url=url,
        throttler_limit_id=limit_id,
        method=RESTMethod.GET,
    )
    server_time = int(response["data"][0]["ts"])

    return server_time


def endpoint_from_message(message: Dict[str, Any]) -> Optional[str]:
    endpoint = None
    if isinstance(message, dict):
        event = message.get("event")
        op = message.get("op")
        if event is not None:
            endpoint = event
        elif op is not None:
            endpoint = op
        else:
            endpoint = message["arg"].get("channel")
    return endpoint


def payload_from_message(message: Dict[str, Any]) -> List[Dict[str, Any]]:
    return message.get("data", [])


def build_api_factory_without_time_synchronizer_pre_processor(throttler: AsyncThrottler) -> WebAssistantsFactory:
    api_factory = WebAssistantsFactory(throttler=throttler)
    return api_factory


def get_rest_url_for_endpoint(
        endpoint: str,
        domain: str = CONSTANTS.DEFAULT_DOMAIN
):
    variant = domain if domain else CONSTANTS.DEFAULT_DOMAIN
    return CONSTANTS.REST_URLS.get(variant) + endpoint


def get_rest_api_limit_id_for_endpoint(method: str, endpoint: str) -> str:
    return f"{method}-{endpoint}"


def get_pair_specific_limit_id(method: str, endpoint: str, trading_pair: str) -> str:
    base_limit_id = get_rest_api_limit_id_for_endpoint(method, endpoint)
    return f"{base_limit_id}-{trading_pair}"


def _wss_url(endpoint: Dict[str, str], connector_variant_label: Optional[str]) -> str:
    variant = connector_variant_label if connector_variant_label else CONSTANTS.DEFAULT_DOMAIN
    return endpoint.get(variant)


def wss_linear_public_url(connector_variant_label: Optional[str]) -> str:
    return _wss_url(CONSTANTS.WSS_PUBLIC_URLS, connector_variant_label)


def wss_linear_private_url(connector_variant_label: Optional[str]) -> str:
    return _wss_url(CONSTANTS.WSS_PRIVATE_URLS, connector_variant_label)


def build_rate_limits(trading_pairs: Optional[List[str]] = None) -> List[RateLimit]:
    trading_pairs = trading_pairs or []
    rate_limits = []
    domain = CONSTANTS.DEFAULT_DOMAIN
    rate_limits.extend(_build_websocket_rate_limits(domain))
    rate_limits.extend(_build_public_rate_limits())
    rate_limits.extend(_build_private_rate_limits(trading_pairs))

    return rate_limits


def _build_websocket_rate_limits(domain: str) -> List[RateLimit]:
    rate_limits = [
        # For connections
        RateLimit(limit_id=CONSTANTS.WSS_PUBLIC_URLS[domain],
                  limit=CONSTANTS.WS_CONNECTION_LIMIT,
                  time_interval=CONSTANTS.ONE_SECOND),
        RateLimit(limit_id=CONSTANTS.WSS_PRIVATE_URLS[domain],
                  limit=CONSTANTS.WS_CONNECTION_LIMIT,
                  time_interval=CONSTANTS.ONE_SECOND),
        # For subscriptions/unsubscriptions/logins
        RateLimit(limit_id=CONSTANTS.WSS_PUBLIC_URLS[domain],
                  limit=CONSTANTS.WS_SUBSCRIPTION_LIMIT,
                  time_interval=CONSTANTS.ONE_MINUTE),
        RateLimit(limit_id=CONSTANTS.WSS_PRIVATE_URLS[domain],
                  limit=CONSTANTS.WS_SUBSCRIPTION_LIMIT,
                  time_interval=CONSTANTS.ONE_MINUTE),
    ]
    return rate_limits


def _build_public_rate_limits():
    public_rate_limits = [
        RateLimit(
            limit_id=get_rest_api_limit_id_for_endpoint(method=CONSTANTS.REST_LATEST_SYMBOL_INFORMATION[CONSTANTS.METHOD],
                                                        endpoint=CONSTANTS.REST_LATEST_SYMBOL_INFORMATION[CONSTANTS.ENDPOINT]),
            limit=CONSTANTS.RATE_LIMIT_LATEST_SYMBOL_INFO,
            time_interval=CONSTANTS.TWO_SECONDS,
        ),
        RateLimit(
            limit_id=get_rest_api_limit_id_for_endpoint(method=CONSTANTS.REST_ORDER_BOOK[CONSTANTS.METHOD],
                                                        endpoint=CONSTANTS.REST_ORDER_BOOK[CONSTANTS.ENDPOINT]),
            limit=CONSTANTS.RATE_LIMIT_ORDER_BOOK,
            time_interval=CONSTANTS.TWO_SECONDS,
        ),
        RateLimit(
            limit_id=get_rest_api_limit_id_for_endpoint(method=CONSTANTS.REST_SERVER_TIME[CONSTANTS.METHOD],
                                                        endpoint=CONSTANTS.REST_SERVER_TIME[CONSTANTS.ENDPOINT]),
            limit=CONSTANTS.RATE_LIMIT_SERVER_TIME,
            time_interval=CONSTANTS.TWO_SECONDS,
        ),
        RateLimit(
            limit_id=get_rest_api_limit_id_for_endpoint(method=CONSTANTS.REST_GET_INSTRUMENTS[CONSTANTS.METHOD],
                                                        endpoint=CONSTANTS.REST_GET_INSTRUMENTS[CONSTANTS.ENDPOINT]),
            limit=CONSTANTS.RATE_LIMIT_GET_INSTRUMENTS,
            time_interval=CONSTANTS.TWO_SECONDS,
        )
    ]
    return public_rate_limits


def _build_private_rate_limits(trading_pairs: List[str]) -> List[RateLimit]:
    rate_limits = []
    rate_limits.extend(_build_private_pair_specific_rate_limits(trading_pairs))
    rate_limits.extend(_build_private_general_rate_limits())
    return rate_limits


def _build_private_pair_specific_rate_limits(trading_pairs: List[str]) -> List[RateLimit]:
    """
    Build pair-specific rate limits for OKX perpetual connector.
    This function is also called when dynamically adding trading pairs.
    """
    rate_limits = []
    for trading_pair in trading_pairs:
        trading_pair_rate_limits = [
            RateLimit(
                limit_id=get_pair_specific_limit_id(method=CONSTANTS.REST_FUNDING_RATE_INFO[CONSTANTS.METHOD],
                                                    endpoint=CONSTANTS.REST_FUNDING_RATE_INFO[CONSTANTS.ENDPOINT],
                                                    trading_pair=trading_pair),
                limit=CONSTANTS.RATE_LIMIT_FUNDING_RATE_INFO,
                time_interval=CONSTANTS.TWO_SECONDS,
            ),
            RateLimit(
                limit_id=get_pair_specific_limit_id(method=CONSTANTS.REST_MARK_PRICE[CONSTANTS.METHOD],
                                                    endpoint=CONSTANTS.REST_MARK_PRICE[CONSTANTS.ENDPOINT],
                                                    trading_pair=trading_pair),
                limit=CONSTANTS.RATE_LIMIT_MARK_PRICE,
                time_interval=CONSTANTS.TWO_SECONDS,
            ),
            RateLimit(
                limit_id=get_pair_specific_limit_id(method=CONSTANTS.REST_INDEX_TICKERS[CONSTANTS.METHOD],
                                                    endpoint=CONSTANTS.REST_INDEX_TICKERS[CONSTANTS.ENDPOINT],
                                                    trading_pair=trading_pair),
                limit=CONSTANTS.RATE_LIMIT_INDEX_TICKERS,
                time_interval=CONSTANTS.TWO_SECONDS,
            ),
        ]
        rate_limits.extend(trading_pair_rate_limits)
    return rate_limits


def _build_private_general_rate_limits() -> List[RateLimit]:
    rate_limits = [
        RateLimit(
            limit_id=get_rest_api_limit_id_for_endpoint(method=CONSTANTS.REST_QUERY_ACTIVE_ORDER[CONSTANTS.METHOD],
                                                        endpoint=CONSTANTS.REST_QUERY_ACTIVE_ORDER[CONSTANTS.ENDPOINT]),
            limit=CONSTANTS.RATE_LIMIT_QUERY_ACTIVE_ORDER,
            time_interval=CONSTANTS.TWO_SECONDS,
        ),
        RateLimit(
            limit_id=get_rest_api_limit_id_for_endpoint(method=CONSTANTS.REST_PLACE_ACTIVE_ORDER[CONSTANTS.METHOD],
                                                        endpoint=CONSTANTS.REST_PLACE_ACTIVE_ORDER[CONSTANTS.ENDPOINT]),
            limit=CONSTANTS.RATE_LIMIT_PLACE_ACTIVE_ORDER,
            time_interval=CONSTANTS.TWO_SECONDS,
        ),
        RateLimit(
            limit_id=get_rest_api_limit_id_for_endpoint(method=CONSTANTS.REST_CANCEL_ACTIVE_ORDER[CONSTANTS.METHOD],
                                                        endpoint=CONSTANTS.REST_CANCEL_ACTIVE_ORDER[CONSTANTS.ENDPOINT]),
            limit=CONSTANTS.RATE_LIMIT_CANCEL_ACTIVE_ORDER,
            time_interval=CONSTANTS.TWO_SECONDS,
        ),
        RateLimit(
            limit_id=get_rest_api_limit_id_for_endpoint(method=CONSTANTS.REST_SET_LEVERAGE[CONSTANTS.METHOD],
                                                        endpoint=CONSTANTS.REST_SET_LEVERAGE[CONSTANTS.ENDPOINT]),
            limit=CONSTANTS.RATE_LIMIT_SET_LEVERAGE,
            time_interval=CONSTANTS.TWO_SECONDS,
        ),
        RateLimit(
            limit_id=get_rest_api_limit_id_for_endpoint(method=CONSTANTS.REST_USER_TRADE_RECORDS[CONSTANTS.METHOD],
                                                        endpoint=CONSTANTS.REST_USER_TRADE_RECORDS[CONSTANTS.ENDPOINT]),
            limit=CONSTANTS.RATE_LIMIT_USER_TRADE_RECORDS,
            time_interval=CONSTANTS.ONE_MINUTE,
        ),
        RateLimit(
            limit_id=get_rest_api_limit_id_for_endpoint(CONSTANTS.REST_GET_POSITIONS[CONSTANTS.METHOD],
                                                        CONSTANTS.REST_GET_POSITIONS[CONSTANTS.ENDPOINT]),
            limit=CONSTANTS.RATE_LIMIT_GET_POSITIONS,
            time_interval=CONSTANTS.TWO_SECONDS,
        ),
        RateLimit(
            limit_id=get_rest_api_limit_id_for_endpoint(method=CONSTANTS.REST_GET_WALLET_BALANCE[CONSTANTS.METHOD],
                                                        endpoint=CONSTANTS.REST_GET_WALLET_BALANCE[CONSTANTS.ENDPOINT]),
            limit=CONSTANTS.RATE_LIMIT_GET_WALLET_BALANCE,
            time_interval=CONSTANTS.TWO_SECONDS,
        ),
        RateLimit(
            limit_id=get_rest_api_limit_id_for_endpoint(method=CONSTANTS.REST_SET_POSITION_MODE[CONSTANTS.METHOD],
                                                        endpoint=CONSTANTS.REST_SET_POSITION_MODE[CONSTANTS.ENDPOINT]),
            limit=CONSTANTS.RATE_LIMIT_SET_POSITION_MODE,
            time_interval=CONSTANTS.TWO_SECONDS,
        ),
        RateLimit(
            limit_id=get_rest_api_limit_id_for_endpoint(method=CONSTANTS.REST_BILLS_DETAILS[CONSTANTS.METHOD],
                                                        endpoint=CONSTANTS.REST_BILLS_DETAILS[CONSTANTS.ENDPOINT]),
            limit=CONSTANTS.RATE_LIMIT_BILLS_DETAILS,
            time_interval=CONSTANTS.ONE_SECOND,
        )
    ]
    return rate_limits



#!/usr/bin/env python

"""
Debug console diagnosis tools.
"""

import asyncio
from typing import Coroutine, Generator, List, Union

import pandas as pd


def get_coro_name(coro: Union[Coroutine, Generator]) -> str:
    if hasattr(coro, '__qualname__') and coro.__qualname__:
        coro_name = coro.__qualname__
    elif hasattr(coro, '__name__') and coro.__name__:
        coro_name = coro.__name__
    else:
        coro_name = f'<{type(coro).__name__} without __name__>'
    return f'{coro_name}()'


def get_wrapped_coroutine(t: asyncio.Task) -> Union[Coroutine, Generator]:
    if "safe_wrapper" in str(t):
        return t.get_coro().cr_frame.f_locals["c"]
    else:
        return t.get_coro()


def active_tasks() -> pd.DataFrame:
    tasks: List[asyncio.Task] = [t for t in asyncio.Task.all_tasks() if not t.done()]
    coroutines: List[Union[Coroutine, Generator]] = [get_wrapped_coroutine(t) for t in tasks]
    func_names: List[str] = [get_coro_name(c) for c in coroutines]
    retval: pd.DataFrame = pd.DataFrame([{"func_name": f, "coroutine": c, "task": t}
                                         for f, c, t in zip(func_names, coroutines, tasks)],
                                        columns=["func_name", "coroutine", "task"]).set_index("func_name")
    retval.sort_index(inplace=True)
    return retval



from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

EXCHANGE_NAME = "binance_perpetual"
BROKER_ID = "x-nbQe1H39"
MAX_ORDER_ID_LEN = 32

DOMAIN = EXCHANGE_NAME
TESTNET_DOMAIN = "binance_perpetual_testnet"

PERPETUAL_BASE_URL = "https://fapi.binance.com/fapi/"
TESTNET_BASE_URL = "https://testnet.binancefuture.com/fapi/"

PERPETUAL_WS_URL = "wss://fstream.binance.com/"
TESTNET_WS_URL = "wss://stream.binancefuture.com/"

PUBLIC_WS_ENDPOINT = "stream"
PRIVATE_WS_ENDPOINT = "ws"

TIME_IN_FORCE_GTC = "GTC"  # Good till cancelled
TIME_IN_FORCE_GTX = "GTX"  # Good Till Crossing
TIME_IN_FORCE_IOC = "IOC"  # Immediate or cancel
TIME_IN_FORCE_FOK = "FOK"  # Fill or kill

# Public API v1 Endpoints
SNAPSHOT_REST_URL = "v1/depth"
TICKER_PRICE_URL = "v1/ticker/bookTicker"
TICKER_PRICE_CHANGE_URL = "v1/ticker/24hr"
EXCHANGE_INFO_URL = "v1/exchangeInfo"
RECENT_TRADES_URL = "v1/trades"
PING_URL = "v1/ping"
MARK_PRICE_URL = "v1/premiumIndex"
SERVER_TIME_PATH_URL = "v1/time"

# Private API v1 Endpoints
ORDER_URL = "v1/order"
CANCEL_ALL_OPEN_ORDERS_URL = "v1/allOpenOrders"
ACCOUNT_TRADE_LIST_URL = "v1/userTrades"
SET_LEVERAGE_URL = "v1/leverage"
GET_INCOME_HISTORY_URL = "v1/income"
CHANGE_POSITION_MODE_URL = "v1/positionSide/dual"

POST_POSITION_MODE_LIMIT_ID = f"POST{CHANGE_POSITION_MODE_URL}"
GET_POSITION_MODE_LIMIT_ID = f"GET{CHANGE_POSITION_MODE_URL}"

# Private API v2 Endpoints
ACCOUNT_INFO_URL = "v2/account"
POSITION_INFORMATION_URL = "v2/positionRisk"

# Private API Endpoints
BINANCE_USER_STREAM_ENDPOINT = "v1/listenKey"

# Funding Settlement Time Span
FUNDING_SETTLEMENT_DURATION = (0, 30)  # seconds before snapshot, seconds after snapshot

# Order Statuses
ORDER_STATE = {
    "NEW": OrderState.OPEN,
    "FILLED": OrderState.FILLED,
    "PARTIALLY_FILLED": OrderState.PARTIALLY_FILLED,
    "CANCELED": OrderState.CANCELED,
    "EXPIRED": OrderState.CANCELED,
    "REJECTED": OrderState.FAILED,
    "EXPIRED_IN_MATCH": OrderState.FAILED,
}

# Rate Limit Type
REQUEST_WEIGHT = "REQUEST_WEIGHT"
ORDERS_1MIN = "ORDERS_1MIN"
ORDERS_1SEC = "ORDERS_1SEC"

DIFF_STREAM_ID = 1
TRADE_STREAM_ID = 2
FUNDING_INFO_STREAM_ID = 3
HEARTBEAT_TIME_INTERVAL = 30.0

# Rate Limit time intervals
ONE_HOUR = 3600
ONE_MINUTE = 60
ONE_SECOND = 1
ONE_DAY = 86400

MAX_REQUEST = 2400

RATE_LIMITS = [
    # Pool Limits
    RateLimit(limit_id=REQUEST_WEIGHT, limit=2400, time_interval=ONE_MINUTE),
    RateLimit(limit_id=ORDERS_1MIN, limit=1200, time_interval=ONE_MINUTE),
    RateLimit(limit_id=ORDERS_1SEC, limit=300, time_interval=10),
    # Weight Limits for individual endpoints
    RateLimit(limit_id=SNAPSHOT_REST_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=20)]),
    RateLimit(limit_id=TICKER_PRICE_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=2)]),
    RateLimit(limit_id=TICKER_PRICE_CHANGE_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=1)]),
    RateLimit(limit_id=EXCHANGE_INFO_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=40)]),
    RateLimit(limit_id=RECENT_TRADES_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=1)]),
    RateLimit(limit_id=BINANCE_USER_STREAM_ENDPOINT, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=1)]),
    RateLimit(limit_id=PING_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=1)]),
    RateLimit(limit_id=SERVER_TIME_PATH_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=1)]),
    RateLimit(limit_id=ORDER_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=1),
                             LinkedLimitWeightPair(ORDERS_1MIN, weight=1),
                             LinkedLimitWeightPair(ORDERS_1SEC, weight=1)]),
    RateLimit(limit_id=CANCEL_ALL_OPEN_ORDERS_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=1)]),
    RateLimit(limit_id=ACCOUNT_TRADE_LIST_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=5)]),
    RateLimit(limit_id=SET_LEVERAGE_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=1)]),
    RateLimit(limit_id=GET_INCOME_HISTORY_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=30)]),
    RateLimit(limit_id=POST_POSITION_MODE_LIMIT_ID, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=1)]),
    RateLimit(limit_id=GET_POSITION_MODE_LIMIT_ID, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=30)]),
    RateLimit(limit_id=ACCOUNT_INFO_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=5)]),
    RateLimit(limit_id=POSITION_INFORMATION_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE, weight=5,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=5)]),
    RateLimit(limit_id=MARK_PRICE_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE, weight=1,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=1)]),
]

ORDER_NOT_EXIST_ERROR_CODE = -2013
ORDER_NOT_EXIST_MESSAGE = "Order does not exist"
UNKNOWN_ORDER_ERROR_CODE = -2011
UNKNOWN_ORDER_MESSAGE = "Unknown order sent"



from eth_account import Account
from eth_account.messages import encode_defunct

from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.connector.utils import to_0x_hex
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTRequest, WSRequest


class DexalotAuth(AuthBase):
    def __init__(self, api_key: str, secret_key: str, time_provider: TimeSynchronizer):
        self.api_key = api_key
        self.secret_key = secret_key
        self.time_provider = time_provider
        self.wallet = Account.from_key(secret_key) if secret_key else None

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.
        :param request: the request to be configured for authenticated interaction
        """

        message = encode_defunct(text="dexalot")
        signed_message = to_0x_hex(self.wallet.sign_message(signable_message=message).signature)

        headers = {"x-signature": f"{self.wallet.address}:{signed_message}"}
        if request.headers is not None:
            headers.update(request.headers)
        request.headers = headers

        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated. Dexalot does not use this
        functionality
        """
        message = encode_defunct(text="dexalot")
        signed_message = to_0x_hex(self.wallet.sign_message(signable_message=message).signature)
        request.payload["signature"] = f"{self.wallet.address}:{signed_message}"
        return request



import asyncio
from typing import TYPE_CHECKING, Dict, Optional

import pandas as pd

from hummingbot.client.config.config_helpers import ClientConfigAdapter
from hummingbot.client.config.security import Security
from hummingbot.client.settings import AllConnectorSettings
from hummingbot.client.ui.interface_utils import format_df_for_printout
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.utils.trading_pair_fetcher import TradingPairFetcher
from hummingbot.user.user_balances import UserBalances

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401

OPTIONS = {cs.name for cs in AllConnectorSettings.get_connector_settings().values()
           if not cs.use_ethereum_wallet and not cs.uses_gateway_generic_connector() if cs.name != "probit_kr"}


class ConnectCommand:
    def connect(self,  # type: HummingbotApplication
                option: str):
        if option is None:
            safe_ensure_future(self.show_connections())
        else:
            safe_ensure_future(self.connect_exchange(option))

    async def connect_exchange(self,  # type: HummingbotApplication
                               connector_name):
        # instruct users to use gateway connect if connector is a gateway connector
        if AllConnectorSettings.get_connector_settings()[connector_name].uses_gateway_generic_connector():
            self.notify("This is a gateway connector. Use `gateway connect` command instead.")
            return

        self.app.clear_input()
        self.placeholder_mode = True
        self.app.hide_input = True
        if connector_name == "kraken":
            self.notify("Reminder: Please ensure your Kraken API Key Nonce Window is at least 10.")
        connector_config = ClientConfigAdapter(AllConnectorSettings.get_connector_config_keys(connector_name))
        if Security.connector_config_file_exists(connector_name):
            await Security.wait_til_decryption_done()
            api_key_config = [value for key, value in Security.api_keys(connector_name).items() if "api_key" in key]
            if api_key_config:
                api_key = api_key_config[0]
                prompt = (
                    f"Would you like to replace your existing {connector_name} API key {api_key} (Yes/No)? >>> "
                )
            else:
                prompt = f"Would you like to replace your existing {connector_name} key (Yes/No)? >>> "
            answer = await self.app.prompt(prompt=prompt)
            if self.app.to_stop_config:
                self.app.to_stop_config = False
                return
            if answer.lower() in ("yes", "y"):
                previous_keys = Security.api_keys(connector_name)
                await self._perform_connect(connector_config, previous_keys)
        else:
            await self._perform_connect(connector_config)
        self.placeholder_mode = False
        self.app.hide_input = False
        self.app.change_prompt(prompt=">>> ")

    async def show_connections(self  # type: HummingbotApplication
                               ):
        self.notify("\nTesting connections, please wait...")
        df, failed_msgs = await self.connection_df()
        lines = ["    " + line for line in format_df_for_printout(
            df,
            table_format=self.client_config_map.tables_format).split("\n")]
        if failed_msgs:
            lines.append("\nFailed connections:")
            lines.extend(["    " + k + ": " + v for k, v in failed_msgs.items()])
        self.notify("\n".join(lines))

    async def connection_df(self  # type: HummingbotApplication
                            ):
        await Security.wait_til_decryption_done()
        columns = ["Exchange", "  Keys Added", "  Keys Confirmed"]
        data = []
        failed_msgs = {}
        network_timeout = float(self.client_config_map.commands_timeout.other_commands_timeout)
        try:
            err_msgs = await asyncio.wait_for(
                UserBalances.instance().update_exchanges(self.client_config_map, reconnect=True), network_timeout
            )
        except asyncio.TimeoutError:
            self.notify("\nA network error prevented the connection table to populate. See logs for more details.")
            raise
        for option in sorted(OPTIONS):
            keys_added = "No"
            keys_confirmed = "No"
            api_keys = (
                Security.api_keys(option).values()
                if not UserBalances.instance().is_gateway_market(option)
                else {}
            )
            if len(api_keys) > 0:
                keys_added = "Yes"
                err_msg = err_msgs.get(option)
                if err_msg is not None:
                    failed_msgs[option] = err_msg
                else:
                    keys_confirmed = "Yes"
            data.append([option, keys_added, keys_confirmed])
        return pd.DataFrame(data=data, columns=columns), failed_msgs

    async def validate_n_connect_connector(
        self,  # type: HummingbotApplication
        connector_name: str,
    ) -> Optional[str]:
        await Security.wait_til_decryption_done()
        api_keys = Security.api_keys(connector_name)
        network_timeout = float(self.client_config_map.commands_timeout.other_commands_timeout)
        try:
            err_msg = await asyncio.wait_for(
                UserBalances.instance().add_exchange(connector_name, self.client_config_map, **api_keys),
                network_timeout,
            )
        except asyncio.TimeoutError:
            self.notify(
                "\nA network error prevented the connection to complete. See logs for more details.")
            self.placeholder_mode = False
            self.app.hide_input = False
            self.app.change_prompt(prompt=">>> ")
            raise
        return err_msg

    async def _perform_connect(self, connector_config: ClientConfigAdapter, previous_keys: Optional[Dict] = None):
        connector_name = connector_config.connector
        original_config = connector_config.full_copy()
        await self.prompt_for_model_config(connector_config)
        self.app.change_prompt(prompt=">>> ")
        if self.app.to_stop_config:
            self.app.to_stop_config = False
            return
        Security.update_secure_config(connector_config)
        err_msg = await self.validate_n_connect_connector(connector_name)
        if err_msg is None:
            self.notify(f"\nYou are now connected to {connector_name}.")
            safe_ensure_future(TradingPairFetcher.get_instance(client_config_map=ClientConfigAdapter).fetch_all(client_config_map=ClientConfigAdapter))
        else:
            self.notify(f"\nError: {err_msg}")
            if previous_keys is not None:
                Security.update_secure_config(original_config)



import asyncio
import logging
from typing import Dict, Optional

import aiohttp

from hummingbot.core.network_base import NetworkBase
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.logger import HummingbotLogger


class DataFeedBase(NetworkBase):
    dfb_logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls.dfb_logger is None:
            cls.dfb_logger = logging.getLogger(__name__)
        return cls.dfb_logger

    def __init__(self):
        super().__init__()
        self._ready_event = asyncio.Event()
        self._shared_client: Optional[aiohttp.ClientSession] = None

    @property
    def name(self):
        raise NotImplementedError

    @property
    def price_dict(self) -> Dict[str, float]:
        raise NotImplementedError

    @property
    def health_check_endpoint(self) -> str:
        raise NotImplementedError

    @property
    def ready(self) -> bool:
        return self._ready_event.is_set()

    def get_price(self, asset: str) -> float:
        raise NotImplementedError

    async def _http_client(self) -> aiohttp.ClientSession:
        if self._shared_client is None:
            self._shared_client = aiohttp.ClientSession()
        return self._shared_client

    async def get_ready(self):
        try:
            if not self._ready_event.is_set():
                await self._ready_event.wait()
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error("Unexpected error while waiting for data feed to get ready.",
                                exc_info=True)

    async def start_network(self):
        raise NotImplementedError

    async def stop_network(self):
        raise NotImplementedError

    async def check_network(self) -> NetworkStatus:
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(self.health_check_endpoint) as resp:
                    status_text = await resp.text()
                    if resp.status != 200:
                        raise Exception(f"Data feed {self.name} server is down. Status is {status_text}")
        except asyncio.CancelledError:
            raise
        except Exception:
            return NetworkStatus.NOT_CONNECTED
        return NetworkStatus.CONNECTED

    def start(self):
        NetworkBase.start(self)

    def stop(self):
        NetworkBase.stop(self)



import errno
import functools
import socket

import cachetools
import numpy as np
import pandas as pd


def async_ttl_cache(ttl: int = 3600, maxsize: int = 1):
    cache = cachetools.TTLCache(ttl=ttl, maxsize=maxsize)

    def decorator(fn):
        @functools.wraps(fn)
        async def memoize(*args, **kwargs):
            key = str((args, kwargs))
            try:
                return cache[key]
            except KeyError:
                cache[key] = await fn(*args, **kwargs)
                return cache[key]

        memoize.cache_clear = lambda: cache.clear()
        return memoize

    return decorator


def map_df_to_str(df: pd.DataFrame) -> pd.DataFrame:
    return df.apply(lambda series: series.map(lambda x: np.format_float_positional(x, trim="-") if isinstance(x, float) else x)).astype(str)


def detect_available_port(starting_port: int) -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        current_port: int = starting_port
        while current_port < 65535:
            try:
                s.bind(("127.0.0.1", current_port))
                break
            except OSError as e:
                if e.errno == errno.EADDRINUSE:
                    current_port += 1
                    continue
        return current_port



#!/usr/bin/env python

from .inventory_cost_price_delegate import InventoryCostPriceDelegate
from .pure_market_making import PureMarketMakingStrategy

__all__ = [
    PureMarketMakingStrategy,
    InventoryCostPriceDelegate,
]



#!/usr/bin/env python
import asyncio
import logging
import time
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

import pandas as pd

from hummingbot.client.command.command_utils import GatewayCommandUtils
from hummingbot.client.command.gateway_api_manager import GatewayChainApiManager, begin_placeholder_mode
from hummingbot.client.config.client_config_map import ClientConfigMap
from hummingbot.client.config.config_helpers import get_connector_class  # noqa: F401
from hummingbot.client.config.security import Security
from hummingbot.client.performance import PerformanceMetrics
from hummingbot.client.settings import AllConnectorSettings, gateway_connector_trading_pairs  # noqa: F401
from hummingbot.client.ui.interface_utils import format_df_for_printout
from hummingbot.core.gateway import get_gateway_paths
from hummingbot.core.gateway.gateway_http_client import GatewayHttpClient, GatewayStatus
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.utils.ssl_cert import create_self_sign_certs

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


def ensure_gateway_online(func):
    def wrapper(self, *args, **kwargs):
        if self.trading_core.gateway_monitor.gateway_status is GatewayStatus.OFFLINE:
            self.logger().error("Gateway is offline")
            return
        return func(self, *args, **kwargs)
    return wrapper


class GatewayCommand(GatewayChainApiManager):
    client_config_map: ClientConfigMap
    _market: Dict[str, Any] = {}

    def __init__(self,  # type: HummingbotApplication
                 client_config_map: ClientConfigMap
                 ):
        super().__init__(client_config_map)
        self.client_config_map = client_config_map

    def gateway(self):
        """Show gateway help when no subcommand is provided."""
        self.notify("""
Gateway Commands:
  gateway allowance <connector> [tokens]                    - Check token allowances
  gateway approve <connector> <tokens>                      - Approve tokens for spending
  gateway balance [chain] [tokens]                          - Check token balances
  gateway config [namespace]                                - Show configuration
  gateway config <namespace> update                         - Update configuration (interactive)
  gateway config <namespace> update <path> <value>          - Update configuration (direct)
  gateway connect <chain>                                   - View and add wallets for a chain
  gateway generate-certs                                    - Generate SSL certificates
  gateway list                                              - List available connectors
  gateway lp <connector> <action>                           - Manage liquidity positions
  gateway ping [chain]                                      - Test node and chain/network status
  gateway pool <connector> <pair>                           - View pool information
  gateway pool <connector> <pair> update                    - Add/update pool information (interactive)
  gateway pool <connector> <pair> update <address>          - Add/update pool information (direct)
  gateway swap <connector> [pair] [side] [amount]           - Swap tokens
  gateway token <symbol_or_address>                         - View token information
  gateway token <symbol> update                             - Update token information

Use 'gateway <command> --help' for more information about a command.""")

    @ensure_gateway_online
    def gateway_status(self):
        safe_ensure_future(self._gateway_status(), loop=self.ev_loop)

    @ensure_gateway_online
    def gateway_balance(self, chain: Optional[str] = None, tokens: Optional[str] = None):
        safe_ensure_future(self._get_balances(chain, tokens), loop=self.ev_loop)

    @ensure_gateway_online
    def gateway_allowance(self, connector: Optional[str] = None):
        """
        Command to check token allowances for Ethereum-based connectors
        Usage: gateway allowance [connector]
        """
        safe_ensure_future(self._get_allowances(connector), loop=self.ev_loop)

    @ensure_gateway_online
    def gateway_approve(self, connector: Optional[str], tokens: Optional[str]):
        # Delegate to GatewayApproveCommand
        from hummingbot.client.command.gateway_approve_command import GatewayApproveCommand
        GatewayApproveCommand.gateway_approve(self, connector, tokens)

    @ensure_gateway_online
    def gateway_connect(self, chain: Optional[str]):
        """
        View and add wallets for a chain.
        Usage: gateway connect <chain>
        """
        if not chain:
            self.notify("\nError: Chain is required")
            self.notify("Usage: gateway connect <chain>")
            self.notify("Example: gateway connect ethereum")
            return

        safe_ensure_future(self._gateway_connect(chain), loop=self.ev_loop)

    def generate_certs(self):
        safe_ensure_future(self._generate_certs(), loop=self.ev_loop)

    @ensure_gateway_online
    def gateway_ping(self, chain: str = None):
        safe_ensure_future(self._gateway_ping(chain), loop=self.ev_loop)

    @ensure_gateway_online
    def gateway_token(self, symbol_or_address: Optional[str], action: Optional[str]):
        # Delegate to GatewayTokenCommand
        from hummingbot.client.command.gateway_token_command import GatewayTokenCommand
        GatewayTokenCommand.gateway_token(self, symbol_or_address, action)

    @ensure_gateway_online
    def gateway_pool(self, connector: Optional[str], trading_pair: Optional[str], action: Optional[str], args: List[str] = None):
        # Delegate to GatewayPoolCommand
        from hummingbot.client.command.gateway_pool_command import GatewayPoolCommand
        GatewayPoolCommand.gateway_pool(self, connector, trading_pair, action, args)

    @ensure_gateway_online
    def gateway_list(self):
        safe_ensure_future(self._gateway_list(), loop=self.ev_loop)

    @ensure_gateway_online
    def gateway_config(self, namespace: str = None, action: str = None, args: List[str] = None):
        # Delegate to GatewayConfigCommand
        from hummingbot.client.command.gateway_config_command import GatewayConfigCommand
        GatewayConfigCommand.gateway_config(self, namespace, action, args)

    async def _gateway_ping(self, chain: str = None):
        """Test gateway connectivity and network status"""
        gateway = self._get_gateway_instance()

        # First test basic gateway connectivity
        if not await gateway.ping_gateway():
            self.notify("\nUnable to ping gateway - gateway service is offline.")
            return

        self.notify("\nGateway service is online.")

        # Get available chains if no specific chain is provided
        if chain is None:
            try:
                chains_resp = await gateway.get_chains()
                if not chains_resp or "chains" not in chains_resp:
                    self.notify("No chains available on gateway.")
                    return

                chains_data = chains_resp["chains"]
                self.notify(f"\nTesting network status for {len(chains_data)} chains...\n")

                # Test each chain with its default network
                for chain_info in chains_data:
                    chain_name = chain_info.get("chain")
                    # Get default network for this chain
                    default_network = await gateway.get_default_network_for_chain(chain_name)
                    if default_network:
                        await self._test_network_status(chain_name, default_network)
                    else:
                        self.notify(f"{chain_name}: No default network configured\n")
            except Exception as e:
                self.notify(f"Error getting chains: {str(e)}")
        else:
            # Test specific chain with its default network
            try:
                # Get default network for the specified chain
                default_network = await gateway.get_default_network_for_chain(chain)
                if default_network:
                    await self._test_network_status(chain, default_network)
                else:
                    self.notify(f"No default network configured for chain: {chain}")
            except Exception as e:
                self.notify(f"Error testing chain {chain}: {str(e)}")

    async def _test_network_status(self, chain: str, network: str):
        """Test network status for a specific chain/network combination"""
        try:
            gateway = self._get_gateway_instance()
            status = await gateway.get_network_status(chain=chain, network=network)

            if status:
                self.notify(f"{chain} ({network}):")
                self.notify(f"  - RPC URL: {status.get('rpcUrl', 'N/A')}")
                self.notify(f"  - Current Block: {status.get('currentBlockNumber', 'N/A')}")
                self.notify(f"  - Native Currency: {status.get('nativeCurrency', 'N/A')}")
                self.notify("  - Status: ✓ Connected\n")
            else:
                self.notify(f"{chain} ({network}): ✗ Unable to get network status\n")
        except Exception as e:
            self.notify(f"{chain} ({network}): ✗ Error - {str(e)}\n")

    async def _gateway_connect(
        self,  # type: HummingbotApplication
        chain: str
    ):
        """View and add wallets for a chain."""
        try:
            # Get default network for the chain
            default_network = await self._get_gateway_instance().get_default_network_for_chain(chain)
            if not default_network:
                self.notify(f"\nError: Could not determine default network for chain '{chain}'")
                self.notify("Please check that the chain name is correct.")
                return

            self.notify(f"\n=== {chain} wallets ===")
            self.notify(f"Network: {default_network}")

            # Get existing wallets to show
            wallets_response = await self._get_gateway_instance().get_wallets(show_hardware=True)

            # Find wallets for this chain
            chain_wallets = None
            for wallet_info in wallets_response:
                if wallet_info.get("chain") == chain:
                    chain_wallets = wallet_info
                    break

            if chain_wallets:
                # Get current default wallet
                default_wallet = await self._get_gateway_instance().get_default_wallet_for_chain(chain)

                # Display existing wallets
                self.notify("\nExisting wallets:")

                # Regular wallets
                wallet_addresses = chain_wallets.get("walletAddresses", [])
                for address in wallet_addresses:
                    if address == default_wallet:
                        self.notify(f"  • {address} (default)")
                    else:
                        self.notify(f"  • {address}")

                    # Check for placeholder wallet
                    if GatewayCommandUtils.is_placeholder_wallet(address):
                        self.notify("    ⚠️  This is a placeholder wallet - please replace it")

                # Hardware wallets
                hardware_addresses = chain_wallets.get("hardwareWalletAddresses", [])
                for address in hardware_addresses:
                    if address == default_wallet:
                        self.notify(f"  • {address} (hardware, default)")
                    else:
                        self.notify(f"  • {address} (hardware)")
            else:
                self.notify("\nNo existing wallets found for this chain.")

            # Enter interactive mode
            with begin_placeholder_mode(self):
                # Ask for wallet type
                wallet_type = await self.app.prompt(
                    prompt="Select Option (1) Add Regular Wallet, (2) Add Hardware Wallet, (3) Exit [default: 3]: "
                )

                if self.app.to_stop_config:
                    self.notify("No wallet added.")
                    return

                # Default to exit if empty input
                if not wallet_type or wallet_type == "3":
                    self.notify("No wallet added.")
                    return

                # Check for valid wallet type
                if wallet_type not in ["1", "2"]:
                    self.notify("Invalid option. No wallet added.")
                    return

                is_hardware = wallet_type == "2"
                wallet_type_str = "hardware" if is_hardware else "regular"

                # For hardware wallets, we need the address instead of private key
                if is_hardware:
                    wallet_input = await self.app.prompt(
                        prompt=f"Enter your {chain} wallet address: "
                    )
                else:
                    wallet_input = await self.app.prompt(
                        prompt=f"Enter your {chain} wallet private key: ",
                        is_password=True
                    )

                if self.app.to_stop_config or not wallet_input:
                    self.notify("Wallet addition cancelled")
                    return

                # Add wallet based on type
                self.notify(f"\nAdding {wallet_type_str} wallet...")

                if is_hardware:
                    # For hardware wallets, pass the address parameter
                    response = await self._get_gateway_instance().add_hardware_wallet(
                        chain=chain,
                        address=wallet_input,  # Hardware wallets use address parameter
                        set_default=True
                    )
                else:
                    # For regular wallets, pass the private key
                    response = await self._get_gateway_instance().add_wallet(
                        chain=chain,
                        private_key=wallet_input,
                        set_default=True
                    )

                # Check response
                if response and "address" in response:
                    self.notify(f"\n✓ Successfully added {wallet_type_str} wallet!")
                    self.notify(f"Address: {response['address']}")
                    self.notify(f"Set as default wallet for {chain}")
                else:
                    error_msg = response.get("error", "Unknown error") if response else "No response"
                    self.notify(f"\n✗ Failed to add wallet: {error_msg}")

        except Exception as e:
            self.notify(f"\nError adding wallet: {str(e)}")
            self.logger().error(f"Error in gateway connect: {e}", exc_info=True)

    async def _generate_certs(
            self,       # type: HummingbotApplication
            from_client_password: bool = False,
    ):

        certs_path: str = get_gateway_paths(
            self.client_config_map).local_certs_path.as_posix()

        if not from_client_password:
            with begin_placeholder_mode(self):
                while True:
                    pass_phase = await self.app.prompt(
                        prompt='Enter pass phrase to generate Gateway SSL certifications  >>> ',
                        is_password=True
                    )
                    if pass_phase is not None and len(pass_phase) > 0:
                        break
                    self.notify("Error: Invalid pass phrase")
        else:
            pass_phase = Security.secrets_manager.password.get_secret_value()
        create_self_sign_certs(pass_phase, certs_path)
        self.notify(
            f"Gateway SSL certification files are created in {certs_path}.")
        self._get_gateway_instance().reload_certs(self.client_config_map.gateway)

    async def ping_gateway_api(self, max_wait: int) -> bool:
        """
        Try to reach the gateway API for up to max_wait seconds
        """
        now = int(time.time())
        gateway_live = await self._get_gateway_instance().ping_gateway()
        while not gateway_live:
            later = int(time.time())
            if later - now > max_wait:
                return False
            await asyncio.sleep(0.5)
            gateway_live = await self._get_gateway_instance().ping_gateway()
            later = int(time.time())

        return True

    async def _gateway_status(self):
        if self.trading_core.gateway_monitor.gateway_status is GatewayStatus.ONLINE:
            try:
                status = await self._get_gateway_instance().get_gateway_status()
                if status is None or status == []:
                    self.notify("There are currently no connectors online.")
                else:
                    self.notify(pd.DataFrame(status))
            except Exception:
                self.notify(
                    "\nError: Unable to fetch status of connected Gateway server.")
        else:
            self.notify(
                "\nNo connection to Gateway server exists. Ensure Gateway server is running.")

    async def _prompt_for_wallet_address(
        self,           # type: HummingbotApplication
        chain: str,
        network: str,
    ) -> Tuple[Optional[str], Dict[str, str]]:
        self.app.clear_input()
        self.placeholder_mode = True
        wallet_private_key = await self.app.prompt(
            prompt=f"Enter your {chain}-{network} wallet private key >>> ",
            is_password=True
        )
        self.app.clear_input()
        if self.app.to_stop_config:
            return

        response: Dict[str, Any] = await self._get_gateway_instance().add_wallet(
            chain, network, wallet_private_key
        )
        wallet_address: str = response["address"]
        return wallet_address

    async def _get_balances(self, chain_filter: Optional[str] = None, tokens_filter: Optional[str] = None):
        network_timeout = float(self.client_config_map.commands_timeout.other_commands_timeout)
        self.notify("Updating gateway balances, please wait...")

        # Determine which chains to check
        chains_to_check = []
        if chain_filter:
            # Check specific chain
            chains_to_check = [chain_filter]
        else:
            # Get all available chains from the Chain enum
            from hummingbot.connector.gateway.common_types import Chain
            chains_to_check = [chain.chain for chain in Chain]

        # Process each chain
        for chain in chains_to_check:
            # Get default network for this chain
            default_network = await self._get_gateway_instance().get_default_network_for_chain(chain)
            if not default_network:
                self.notify(f"Could not determine default network for {chain}")
                continue

            # Get default wallet for this chain
            default_wallet = await self._get_gateway_instance().get_default_wallet_for_chain(chain)
            if not default_wallet:
                self.notify(f"No default wallet found for {chain}. Please add one with 'gateway connect {chain}'")
                continue

            # Check if wallet address is a placeholder
            if GatewayCommandUtils.is_placeholder_wallet(default_wallet):
                self.notify(f"\n⚠️  {chain} wallet not configured (found placeholder: {default_wallet})")
                self.notify(f"Please add a real wallet with: gateway connect {chain}")
                continue

            try:
                # Determine tokens to check
                if tokens_filter:
                    # User specified tokens (comma-separated)
                    tokens_to_check = [token.strip() for token in tokens_filter.split(",")]

                    # Validate tokens
                    valid_tokens, invalid_tokens = await self._get_gateway_instance().validate_tokens(
                        chain, default_network, tokens_to_check
                    )

                    if invalid_tokens:
                        self.notify(f"\n❌ Unknown tokens for {chain}: {', '.join(invalid_tokens)}")
                        self.notify("Please check the token symbol(s) and try again.")
                        continue

                    # Use validated tokens
                    tokens_to_check = valid_tokens
                else:
                    # No filter specified - fetch all tokens
                    tokens_to_check = []

                # Get balances from gateway
                tokens_display = "all" if not tokens_to_check else ", ".join(tokens_to_check)
                self.notify(f"\nFetching balances for {chain}:{default_network} for tokens: {tokens_display}")
                balances_resp = await asyncio.wait_for(
                    self._get_gateway_instance().get_balances(chain, default_network, default_wallet, tokens_to_check),
                    network_timeout
                )
                balances = balances_resp.get("balances", {})

                # Show all balances including zero balances
                display_balances = balances

                # Display results
                self.notify(f"\nChain: {chain.lower()}")
                self.notify(f"Network: {default_network}")
                self.notify(f"Address: {default_wallet}")

                if display_balances:
                    rows = []
                    for token, bal in display_balances.items():
                        rows.append({
                            "Token": token.upper(),
                            "Balance": PerformanceMetrics.smart_round(Decimal(str(bal)), 4),
                        })

                    df = pd.DataFrame(data=rows, columns=["Token", "Balance"])
                    df.sort_values(by=["Token"], inplace=True)

                    lines = [
                        "    " + line for line in df.to_string(index=False).split("\n")
                    ]
                    self.notify("\n".join(lines))
                else:
                    self.notify("    No balances found")

            except asyncio.TimeoutError:
                self.notify(f"\nError getting balance for {chain}:{default_network}: Request timed out")

    @staticmethod
    async def _update_balances(market) -> Optional[str]:
        try:
            await market._update_balances()
        except Exception as e:
            logging.getLogger().debug(
                f"Failed to update balances for {market}", exc_info=True)
            return str(e)
        return None

    def all_balance(self, exchange) -> Dict[str, Decimal]:
        if exchange not in self._market:
            return {}
        return self._market[exchange].get_all_balances()

    async def update_exchange(
        self,
        client_config_map: ClientConfigMap,
        reconnect: bool = False,
        exchanges: Optional[List[str]] = None
    ) -> Dict[str, Optional[str]]:
        """
        Simple gateway balance update for compatibility.
        Returns empty dict (no errors) since gateway balances are fetched on-demand.
        """
        # Gateway balances are fetched directly from the gateway when needed
        # No need to maintain cached balances like CEX connectors
        return {}

    async def balance(self, exchange, client_config_map: ClientConfigMap, *symbols) -> Dict[str, Decimal]:
        """
        Get balances for specified tokens from a gateway connector.

        Args:
            exchange: The gateway connector name (e.g., "uniswap_ethereum_mainnet")
            client_config_map: Client configuration
            *symbols: Token symbols to get balances for

        Returns:
            Dict mapping token symbols to their balances
        """
        try:
            # Parse exchange name to get connector format
            # Exchange names like "uniswap_ethereum_mainnet" need to be converted to "uniswap/amm" format
            parts = exchange.split("_")
            if len(parts) < 1:
                self.logger().warning(f"Invalid gateway exchange format: {exchange}")
                return {}

            # The connector name is the first part
            connector_name = parts[0]

            # Determine connector type - this is a simplified mapping
            # In practice, this should be determined from the connector settings
            connector_type = "amm"  # Default to AMM for now
            connector = f"{connector_name}/{connector_type}"

            # Get chain and network from the connector
            gateway = self._get_gateway_instance()
            chain, network, error = await self._get_gateway_instance().get_connector_chain_network(
                connector
            )

            if error:
                self.logger().warning(f"Error getting chain/network for {exchange}: {error}")
                return {}

            # Get default wallet for the chain
            default_wallet = await gateway.get_default_wallet_for_chain(chain)
            if not default_wallet:
                self.logger().warning(f"No default wallet for chain {chain}")
                return {}

            # Fetch balances directly from gateway
            tokens_list = list(symbols) if symbols else []
            balances_resp = await gateway.get_balances(chain, network, default_wallet, tokens_list)
            balances = balances_resp.get("balances", {})

            # Convert to Decimal and match requested symbols
            results = {}
            for token, balance in balances.items():
                for symbol in symbols:
                    if token.lower() == symbol.lower():
                        results[symbol] = Decimal(str(balance))
                        break

            return results

        except Exception as e:
            self.logger().error(f"Error getting gateway balances: {e}", exc_info=True)
            return {}

    async def _gateway_list(
        self           # type: HummingbotApplication
    ):
        connector_list: List[Dict[str, Any]] = await self._get_gateway_instance().get_connectors()
        connectors_tiers: List[Dict[str, Any]] = []

        for connector in connector_list["connectors"]:
            # Chain and networks are now directly in the connector config
            chain = connector["chain"]
            networks = connector["networks"]

            # Convert to string for display
            chain_type_str = chain
            networks_str = ", ".join(networks) if networks else "N/A"

            # Extract trading types and convert to string
            trading_types: List[str] = connector.get("trading_types", [])
            trading_types_str = ", ".join(trading_types) if trading_types else "N/A"

            # Create a new dictionary with the fields we want to display
            display_connector = {
                "connector": connector.get("name", ""),
                "chain_type": chain_type_str,  # Use string instead of list
                "networks": networks_str,      # Use string instead of list
                "trading_types": trading_types_str
            }

            connectors_tiers.append(display_connector)

        # Make sure to include all fields in the dataframe
        columns = ["connector", "chain_type", "networks", "trading_types"]
        connectors_df = pd.DataFrame(connectors_tiers, columns=columns)

        lines = ["    " + line for line in format_df_for_printout(
            connectors_df,
            table_format=self.client_config_map.tables_format).split("\n")]
        self.notify("\n".join(lines))

    def _get_gateway_instance(
        self  # type: HummingbotApplication
    ) -> GatewayHttpClient:
        gateway_instance = GatewayHttpClient.get_instance(
            self.client_config_map)
        return gateway_instance

    async def _get_allowances(self, connector: Optional[str] = None):
        """Get token allowances for Ethereum-based connectors"""
        gateway_instance = self._get_gateway_instance()
        self.notify("Checking token allowances, please wait...")

        try:
            # If specific connector requested
            if connector is not None:
                # Parse connector format (e.g., "uniswap/amm")
                if "/" not in connector:
                    self.notify(f"Error: Invalid connector format '{connector}'. Use format like 'uniswap/amm'")
                    return

                # Get chain and network from connector
                chain, network, error = await self._get_gateway_instance().get_connector_chain_network(
                    connector
                )
                if error:
                    self.notify(error)
                    return

                if chain.lower() != "ethereum":
                    self.notify(f"Allowances are only applicable for Ethereum chains. {connector} uses {chain}.")
                    return

                # Get default wallet
                wallet_address, error = await gateway_instance.get_default_wallet(chain)
                if error:
                    self.notify(error)
                    return

                # Get all available tokens for this chain/network
                token_list = await self._get_gateway_instance().get_available_tokens(chain, network)
                if not token_list:
                    self.notify(f"No tokens found for {chain}:{network}")
                    return

                # Create a dict of token symbol to token info
                token_data = {token["symbol"]: token for token in token_list}
                token_symbols = [token["symbol"] for token in token_list]

                # Get allowances using connector including trading type (spender in Gateway)
                allowance_resp = await gateway_instance.get_allowances(
                    chain, network, wallet_address, token_symbols, connector, fail_silently=True
                )

                # Format allowances using the helper
                if allowance_resp.get("approvals") is not None:
                    rows = GatewayCommandUtils.format_allowance_display(
                        allowance_resp["approvals"],
                        token_data=token_data
                    )
                else:
                    rows = []

                if rows:
                    # We always have address data now
                    columns = ["Symbol", "Address", "Allowance"]
                    df = pd.DataFrame(data=rows, columns=columns)
                    df.sort_values(by=["Symbol"], inplace=True)
                else:
                    df = pd.DataFrame()

                # Display connector with spender address in parentheses
                spender_display = ""
                if allowance_resp.get("spender"):
                    spender = allowance_resp["spender"]
                    formatted_spender = f"{spender[:6]}...{spender[-4:]}" if len(spender) > 12 else spender
                    spender_display = f" ({formatted_spender})"

                self.notify(f"\nConnector: {connector}{spender_display}")
                self.notify(f"Chain: {chain}")
                self.notify(f"Network: {network}")
                self.notify(f"Wallet: {wallet_address}")

                if df.empty:
                    self.notify("No token allowances found.")
                else:
                    lines = [
                        "    " + line for line in df.to_string(index=False).split("\n")
                    ]
                    self.notify("\n".join(lines))
            else:
                # Show allowances for all Ethereum connectors
                self.notify("Checking allowances for all Ethereum-based connectors...")

                # Get all connectors
                connectors_resp = await gateway_instance.get_connectors()
                if "error" in connectors_resp:
                    self.notify(f"Error getting connectors: {connectors_resp['error']}")
                    return

                ethereum_connectors = []
                for conn in connectors_resp.get("connectors", []):
                    if conn.get("chain", "").lower() == "ethereum":
                        # Get trading types for this connector
                        trading_types = conn.get("trading_types", [])
                        for trading_type in trading_types:
                            ethereum_connectors.append(f"{conn['name']}/{trading_type}")

                if not ethereum_connectors:
                    self.notify("No Ethereum-based connectors found.")
                    return

                # Get allowances for each ethereum connector
                for connector_name in ethereum_connectors:
                    await self._get_allowances(connector_name)

        except asyncio.TimeoutError:
            self.notify("\nA network error prevented the allowances from updating. See logs for more details.")
            raise
        except Exception as e:
            self.notify(f"\nError getting allowances: {str(e)}")
            self.logger().error(f"Error getting allowances: {e}", exc_info=True)



from decimal import Decimal

from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
from hummingbot.strategy.spot_perpetual_arbitrage.spot_perpetual_arbitrage import SpotPerpetualArbitrageStrategy
from hummingbot.strategy.spot_perpetual_arbitrage.spot_perpetual_arbitrage_config_map import (
    spot_perpetual_arbitrage_config_map,
)


async def start(self):
    spot_connector = spot_perpetual_arbitrage_config_map.get("spot_connector").value.lower()
    spot_market = spot_perpetual_arbitrage_config_map.get("spot_market").value
    perpetual_connector = spot_perpetual_arbitrage_config_map.get("perpetual_connector").value.lower()
    perpetual_market = spot_perpetual_arbitrage_config_map.get("perpetual_market").value
    order_amount = spot_perpetual_arbitrage_config_map.get("order_amount").value
    perpetual_leverage = spot_perpetual_arbitrage_config_map.get("perpetual_leverage").value
    min_opening_arbitrage_pct = spot_perpetual_arbitrage_config_map.get("min_opening_arbitrage_pct").value / Decimal("100")
    min_closing_arbitrage_pct = spot_perpetual_arbitrage_config_map.get("min_closing_arbitrage_pct").value / Decimal("100")
    spot_market_slippage_buffer = spot_perpetual_arbitrage_config_map.get("spot_market_slippage_buffer").value / Decimal("100")
    perpetual_market_slippage_buffer = spot_perpetual_arbitrage_config_map.get("perpetual_market_slippage_buffer").value / Decimal("100")
    next_arbitrage_opening_delay = spot_perpetual_arbitrage_config_map.get("next_arbitrage_opening_delay").value

    await self.initialize_markets([(spot_connector, [spot_market]), (perpetual_connector, [perpetual_market])])
    base_1, quote_1 = spot_market.split("-")
    base_2, quote_2 = perpetual_market.split("-")

    spot_market_info = MarketTradingPairTuple(self.markets[spot_connector], spot_market, base_1, quote_1)
    perpetual_market_info = MarketTradingPairTuple(self.markets[perpetual_connector], perpetual_market, base_2, quote_2)

    self.market_trading_pair_tuples = [spot_market_info, perpetual_market_info]
    self.strategy = SpotPerpetualArbitrageStrategy()
    self.strategy.init_params(spot_market_info,
                              perpetual_market_info,
                              order_amount,
                              perpetual_leverage,
                              min_opening_arbitrage_pct,
                              min_closing_arbitrage_pct,
                              spot_market_slippage_buffer,
                              perpetual_market_slippage_buffer,
                              next_arbitrage_opening_delay)



#!/usr/bin/env python

import asyncio
import base64
import logging
import pickle
import time
from typing import AsyncIterable, Dict, Optional, Tuple

import aiohttp
import pandas as pd
import websockets
from websockets.exceptions import ConnectionClosed

import conf
from hummingbot.connector.exchange.binance.binance_order_book import BinanceOrderBook
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.order_book_tracker_entry import OrderBookTrackerEntry
from hummingbot.logger import HummingbotLogger


class RemoteAPIOrderBookDataSource(OrderBookTrackerDataSource):
    SNAPSHOT_REST_URL = "https://api.coinalpha.com/order_book_tracker/snapshot"
    SNAPSHOT_STREAM_URL = "wss://api.coinalpha.com/ws/order_book_tracker/snapshot_stream"
    DIFF_STREAM_URL = "wss://api.coinalpha.com/ws/order_book_tracker/diff_stream"

    MESSAGE_TIMEOUT = 30.0
    PING_TIMEOUT = 10.0

    _raobds_logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._raobds_logger is None:
            cls._raobds_logger = logging.getLogger(__name__)
        return cls._raobds_logger

    def __init__(self):
        super().__init__()
        self._client_session: Optional[aiohttp.ClientSession] = None

    @property
    def authentication_headers(self) -> Dict[str, str]:
        auth_str: str = f"{conf.coinalpha_order_book_api_username}:{conf.coinalpha_order_book_api_password}"
        encoded_auth: str = base64.standard_b64encode(auth_str.encode("utf8")).decode("utf8")
        return {
            "Authorization": f"Basic {encoded_auth}"
        }

    async def get_client_session(self) -> aiohttp.ClientSession:
        if self._client_session is None:
            self._client_session = aiohttp.ClientSession()
        return self._client_session

    async def get_tracking_pairs(self) -> Dict[str, OrderBookTrackerEntry]:
        auth: aiohttp.BasicAuth = aiohttp.BasicAuth(login=conf.coinalpha_order_book_api_username,
                                                    password=conf.coinalpha_order_book_api_password)
        client_session: aiohttp.ClientSession = await self.get_client_session()
        response: aiohttp.ClientResponse = await client_session.get(self.SNAPSHOT_REST_URL, auth=auth)
        timestamp: float = time.time()
        if response.status != 200:
            raise EnvironmentError(f"Error fetching order book tracker snapshot from {self.SNAPSHOT_REST_URL}.")

        binary_data: bytes = await response.read()
        order_book_tracker_data: Dict[str, Tuple[pd.DataFrame, pd.DataFrame]] = pickle.loads(binary_data)
        retval: Dict[str, OrderBookTrackerEntry] = {}

        for trading_pair, (bids_df, asks_df) in order_book_tracker_data.items():
            order_book: BinanceOrderBook = BinanceOrderBook()
            order_book.apply_numpy_snapshot(bids_df.values, asks_df.values)
            retval[trading_pair] = OrderBookTrackerEntry(trading_pair, timestamp, order_book)

        return retval

    async def _inner_messages(self,
                              ws: websockets.WebSocketClientProtocol) -> AsyncIterable[str]:
        # Terminate the recv() loop as soon as the next message timed out, so the outer loop can reconnect.
        try:
            while True:
                try:
                    msg: str = await asyncio.wait_for(ws.recv(), timeout=self.MESSAGE_TIMEOUT)
                    yield msg
                except asyncio.TimeoutError:
                    pong_waiter = await ws.ping()
                    await asyncio.wait_for(pong_waiter, timeout=self.PING_TIMEOUT)
        except asyncio.TimeoutError:
            self.logger().warning("WebSocket ping timed out. Going to reconnect...")
            return
        except ConnectionClosed:
            return
        finally:
            await ws.close()

    async def listen_for_order_book_diffs(self, ev_loop: asyncio.BaseEventLoop, output: asyncio.Queue):
        while True:
            try:
                async with websockets.connect(self.DIFF_STREAM_URL,
                                              extra_headers=self.authentication_headers) as ws:
                    ws: websockets.WebSocketClientProtocol = ws
                    async for msg in self._inner_messages(ws):
                        output.put_nowait(msg)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error("Unexpected error with WebSocket connection. Retrying after 30 seconds...",
                                    exc_info=True)
                await asyncio.sleep(30.0)

    async def listen_for_order_book_snapshots(self, ev_loop: asyncio.BaseEventLoop, output: asyncio.Queue):
        while True:
            try:
                async with websockets.connect(self.SNAPSHOT_STREAM_URL,
                                              extra_headers=self.authentication_headers) as ws:
                    ws: websockets.WebSocketClientProtocol = ws
                    async for msg in self._inner_messages(ws):
                        output.put_nowait(msg)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error("Unexpected error with WebSocket connection. Retrying after 30 seconds...",
                                    exc_info=True)
                await asyncio.sleep(30.0)
